from __future__ import annotations

from enum import IntEnum


class RngCallerStatic(IntEnum):
    BONUS_PICK_RANDOM_TYPE_ROLL = 0x004124A5
    BONUS_PICK_RANDOM_TYPE_ENERGIZER = 0x004124D6
    BONUS_SPAWN_AT_POS_POINTS_AMOUNT = 0x0041F897
    BONUS_APPLY_NUKE_BULLET_COUNT = 0x0040A00B
    BONUS_APPLY_NUKE_PISTOL_ANGLE = 0x0040A023
    BONUS_APPLY_NUKE_PISTOL_SPEED_SCALE = 0x0040A059
    BONUS_APPLY_NUKE_GAUSS_ANGLE_1 = 0x0040A088
    BONUS_APPLY_NUKE_GAUSS_ANGLE_2 = 0x0040A0B4
    BONUS_TRY_SPAWN_ON_KILL_PISTOL_FORCE_WEAPON = 0x0041F92C
    BONUS_TRY_SPAWN_ON_KILL_BASE_GATE = 0x0041F9AA
    BONUS_TRY_SPAWN_ON_KILL_PISTOL_ALLOW_WITHOUT_MAGNET = 0x0041F9C3
    BONUS_TRY_SPAWN_ON_KILL_BONUS_MAGNET = 0x0041F9EB
    BONUS_TRY_SPAWN_ON_KILL_BURST_ROTATION = 0x0041FB0F
    BONUS_TRY_SPAWN_ON_KILL_BURST_VEL_X = 0x0041FB2B
    BONUS_TRY_SPAWN_ON_KILL_BURST_VEL_Y = 0x0041FB4D
    BONUS_TRY_SPAWN_ON_KILL_BURST_SCALE_STEP = 0x0041FB6F

    PLAYER_UPDATE_LOW_HEALTH_BLOODSPILL = 0x0041381C
    PLAYER_UPDATE_CASING_ANGLE = 0x00415A6D
    PLAYER_UPDATE_CASING_SPEED = 0x00415A8B
    PLAYER_UPDATE_CASING_ROTATION = 0x00415B26
    PLAYER_UPDATE_CASING_ROTATION_STEP = 0x00415B79
    PLAYER_UPDATE_SHOT_JITTER_DIR = 0x00415C6B
    PLAYER_UPDATE_SHOT_JITTER_MAG = 0x00415C87
    PLAYER_UPDATE_SHOT_SFX = 0x00415F09
    PLAYER_UPDATE_SHOTGUN_PELLET_JITTER = 0x00416364
    PLAYER_UPDATE_SHOTGUN_PELLET_SPEED_SCALE = 0x00416399
    PLAYER_UPDATE_SAWED_OFF_SHOTGUN_PELLET_JITTER = 0x00416478
    PLAYER_UPDATE_SAWED_OFF_SHOTGUN_PELLET_SPEED_SCALE = 0x004164AD
    PLAYER_UPDATE_JACKHAMMER_PELLET_JITTER = 0x004165FE
    PLAYER_UPDATE_JACKHAMMER_PELLET_SPEED_SCALE = 0x00416633
    PLAYER_UPDATE_ION_SHOTGUN_PELLET_JITTER = 0x00416B44
    PLAYER_UPDATE_ION_SHOTGUN_PELLET_SPEED_SCALE = 0x00416B79
    PLAYER_UPDATE_GAUSS_SHOTGUN_PELLET_JITTER = 0x00416CFD
    PLAYER_UPDATE_GAUSS_SHOTGUN_PELLET_SPEED_SCALE = 0x00416D32
    PLAYER_UPDATE_PLASMA_SHOTGUN_PELLET_JITTER = 0x0041735A
    PLAYER_UPDATE_PLASMA_SHOTGUN_PELLET_SPEED_SCALE = 0x0041738E
    PLAYER_UPDATE_FIRE_COUGH_SPREAD_DIR = 0x00413B16
    PLAYER_UPDATE_FIRE_COUGH_SPREAD_MAG = 0x00413B32
    PLAYER_UPDATE_FIRE_COUGH_INTERVAL_RESET = 0x00413C4F
    PLAYER_UPDATE_HOT_TEMPERED_INTERVAL_RESET = 0x00413D3A
    PLAYER_UPDATE_MAN_BOMB_ION_RIFLE_ANGLE = 0x00413924
    PLAYER_UPDATE_MAN_BOMB_ION_MINIGUN_ANGLE = 0x0041393D

    CREATURE_ALLOC_SLOT_PHASE_SEED = 0x004281A2
    SURVIVAL_SPAWN_CREATURE_TYPE_ROLL = 0x0040754B
    SURVIVAL_SPAWN_CREATURE_PARITY_PICK = 0x004075B8
    SURVIVAL_SPAWN_CREATURE_RARE_OVERRIDE = 0x00407616
    SURVIVAL_SPAWN_CREATURE_SIZE = 0x00407625
    SURVIVAL_SPAWN_CREATURE_HEADING = 0x0040766C
    SURVIVAL_SPAWN_CREATURE_HEALTH = 0x004076D9
    SURVIVAL_SPAWN_CREATURE_LOW_TINT_G = 0x00407792
    SURVIVAL_SPAWN_CREATURE_LOW_TINT_B = 0x004077E3
    SURVIVAL_SPAWN_CREATURE_MID_TINT_G = 0x00407848
    SURVIVAL_SPAWN_CREATURE_MID_TINT_B = 0x00407899
    SURVIVAL_SPAWN_CREATURE_HIGH_TINT_G = 0x00407906
    SURVIVAL_SPAWN_CREATURE_HIGH_TINT_B = 0x00407957
    SURVIVAL_SPAWN_CREATURE_REWARD_BONUS = 0x004079CF
    SURVIVAL_SPAWN_CREATURE_RARE_RED = 0x00407A0A
    SURVIVAL_SPAWN_CREATURE_RARE_GREEN = 0x00407A6C
    SURVIVAL_SPAWN_CREATURE_RARE_BLUE = 0x00407ACB
    SURVIVAL_SPAWN_CREATURE_RARE_PURPLE = 0x00407B28
    SURVIVAL_SPAWN_CREATURE_RARE_YELLOW = 0x00407B93
    SURVIVAL_UPDATE_EXTRA_SPAWN_EDGE = 0x00408338
    SURVIVAL_UPDATE_EXTRA_SPAWN_TOP_X = 0x00408356
    SURVIVAL_UPDATE_EXTRA_SPAWN_BOTTOM_X = 0x00408381
    SURVIVAL_UPDATE_EXTRA_SPAWN_LEFT_Y = 0x004083B2
    SURVIVAL_UPDATE_EXTRA_SPAWN_RIGHT_Y = 0x004083DB
    SURVIVAL_UPDATE_MAIN_SPAWN_EDGE = 0x00408423
    SURVIVAL_UPDATE_MAIN_SPAWN_TOP_X = 0x00408441
    SURVIVAL_UPDATE_MAIN_SPAWN_BOTTOM_X = 0x0040846C
    SURVIVAL_UPDATE_MAIN_SPAWN_LEFT_Y = 0x0040849D
    SURVIVAL_UPDATE_MAIN_SPAWN_RIGHT_Y = 0x004084C6
    CREATURE_SPAWN_HEADING = 0x004282DC
    CREATURE_SPAWN_REWARD = 0x00428315
    CREATURE_HANDLE_DEATH_SPLIT_CHILD_1_PHASE_SEED = 0x0041E9F4
    CREATURE_HANDLE_DEATH_SPLIT_CHILD_2_PHASE_SEED = 0x0041EA91
    CREATURE_HANDLE_DEATH_FREEZE_SHARD_ANGLE = 0x0041EBEA
    CREATURE_HANDLE_DEATH_FREEZE_SHATTER_ANGLE = 0x0041EC15
    CREATURE_SPAWN_TEMPLATE_RANDOM_HEADING = 0x00430B14
    CREATURE_SPAWN_TEMPLATE_BASE_HEADING = 0x00430B88
    CREATURE_SPAWN_TEMPLATE_AI1_BLUE_TINT_1A = 0x004323EC
    CREATURE_SPAWN_TEMPLATE_AI1_BLUE_TINT_1B = 0x00432440
    CREATURE_SPAWN_TEMPLATE_AI1_BLUE_TINT_1C = 0x0043248B
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_SIZE = 0x0043303B
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_MOVE_SPEED = 0x00433065
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_REWARD = 0x00433089
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_TINT_R = 0x004330AB
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_TINT_G = 0x004330CF
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_TINT_B = 0x004330F3
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1D_CONTACT_DAMAGE = 0x00433117
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_SIZE = 0x0043314A
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_MOVE_SPEED = 0x00433174
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_REWARD = 0x00433198
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_TINT_R = 0x004331BA
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_TINT_G = 0x004331DE
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_TINT_B = 0x00433202
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1E_CONTACT_DAMAGE = 0x00433226
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_SIZE = 0x00433259
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_MOVE_SPEED = 0x00433283
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_REWARD = 0x004332A7
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_TINT_R = 0x004332C9
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_TINT_G = 0x004332ED
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_TINT_B = 0x00433311
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_1F_CONTACT_DAMAGE = 0x00433335
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_GREEN_20_SIZE = 0x00432857
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_GREEN_20_MOVE_SPEED = 0x00432881
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_GREEN_20_TINT_G = 0x004328C2
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_GREEN_20_CONTACT_DAMAGE = 0x004328E9
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_03_SIZE = 0x0043291C
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_03_MOVE_SPEED = 0x00432946
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_03_TINT_B = 0x0043298B
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_03_CONTACT_DAMAGE = 0x00432A4F
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_05_SIZE = 0x00432A82
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_05_MOVE_SPEED = 0x00432AAC
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_05_TINT_B = 0x00432AF1
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_05_CONTACT_DAMAGE = 0x00432BB5
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_04_SIZE = 0x00432BE8
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_04_MOVE_SPEED = 0x00432C12
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_04_CONTACT_DAMAGE = 0x00432C5A
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_06_SIZE = 0x00432C8D
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_06_MOVE_SPEED = 0x00432CB7
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_06_TINT_B = 0x00432CFC
    CREATURE_SPAWN_TEMPLATE_ALIEN_RANDOM_06_CONTACT_DAMAGE = 0x00432DC0
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_31_SIZE = 0x00432579
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_31_MOVE_SPEED = 0x004325A3
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_31_TINT = 0x004325DC
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_32_SIZE = 0x0043261F
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_32_MOVE_SPEED = 0x00432643
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_32_TINT = 0x0043267C
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_RED_33_SIZE = 0x004326CD
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_RED_33_MOVE_SPEED = 0x004326F7
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_RED_33_TINT_R = 0x00432730
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_RED_33_CONTACT_DAMAGE = 0x0043275F
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_GREEN_34_SIZE = 0x00432792
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_GREEN_34_MOVE_SPEED = 0x004327BC
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_GREEN_34_TINT_G = 0x004327F5
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_GREEN_34_CONTACT_DAMAGE = 0x00432824
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_2E_SIZE = 0x00432EB8
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_2E_MOVE_SPEED = 0x00432EE2
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_2E_TINT_R = 0x00432F1B
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_2E_TINT_G = 0x00432F3F
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_2E_TINT_B = 0x00432F63
    CREATURE_SPAWN_TEMPLATE_LIZARD_RANDOM_2E_CONTACT_DAMAGE = 0x00432F87
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_35_SIZE = 0x00432DF3
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_35_MOVE_SPEED = 0x00432E1D
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_35_TINT_G = 0x00432E5E
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANDOM_35_CONTACT_DAMAGE = 0x00432E85
    CREATURE_SPAWN_TEMPLATE_AI7_ORBITER_TINT_G = 0x00432FED
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_3D_TINT = 0x00433B39
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_RANDOM_3D_SIZE = 0x00433B63
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_AI7_TIMER_38_SIZE = 0x00433D18
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP2_RANGED_VARIANT_37_SIZE = 0x00433DB3
    CREATURE_SPAWN_TEMPLATE_SPIDER_SP1_AI7_TIMER_WEAK_39_SIZE = 0x00433E4E
    CREATURE_SPAWN_TEMPLATE_ZOMBIE_RANDOM_41_SIZE = 0x004324CD
    CREATURE_SPAWN_TEMPLATE_ZOMBIE_RANDOM_41_TINT = 0x0043251C
    CREATURE_SPAWN_TEMPLATE_ZOMBIE_RANDOM_41_CONTACT_DAMAGE = 0x00432546

    CREATURE_UPDATE_ALL_AI7_LINK_TIMER_HOLD = 0x00426381
    CREATURE_UPDATE_ALL_AI7_LINK_TIMER_RESET = 0x004263B1
    CREATURE_UPDATE_ALL_CONTACT_SFX = 0x00426611
    CREATURE_UPDATE_ALL_PLASMAMINIGUN_COOLDOWN = 0x00427119
    CREATURE_UPDATE_ALL_PLAGUE_KILL_SFX = 0x00427273
    CREATURE_UPDATE_ALL_PING_PONG_BLOOD_8_ANGLE = 0x004275A3
    CREATURE_UPDATE_ALL_PING_PONG_BLOOD_6_ANGLE = 0x004275D8
    CREATURE_UPDATE_ALL_PING_PONG_BLOOD_5_ANGLE = 0x0042760D

    CREATURE_APPLY_DAMAGE_HEADING_JITTER = 0x004208A9
    CREATURE_APPLY_DAMAGE_PYROMANIAC = 0x0042092D
    CREATURE_APPLY_DAMAGE_DEATH_SFX = 0x00420B16
    CREATURE_APPLY_DAMAGE_SHOCK_BURST_ROTATION = 0x00420A4B
    CREATURE_APPLY_DAMAGE_SHOCK_BURST_VEL_X = 0x00420A67
    CREATURE_APPLY_DAMAGE_SHOCK_BURST_VEL_Y = 0x00420A80
    CREATURE_APPLY_DAMAGE_SHOCK_BURST_SCALE_STEP = 0x00420A99

    BONUS_APPLY_FREEZE_SHARD_ANGLE = 0x00409BDC
    BONUS_APPLY_FREEZE_SHATTER_ANGLE = 0x00409C07

    PROJECTILE_UPDATE_POISON_BULLETS_GATE = 0x00420E84
    PROJECTILE_UPDATE_BLADE_GUN_SPLATTER_ANGLE = 0x00420EAF
    PROJECTILE_UPDATE_BLOODY_MESS_SPREAD = 0x00420FD7
    PROJECTILE_UPDATE_DEFAULT_REVERSE_SPLATTER_GATE = 0x0042105B
    PROJECTILE_UPDATE_BLOODY_MESS_DECAL_DX_1 = 0x004210D3
    PROJECTILE_UPDATE_BLOODY_MESS_DECAL_DY_1 = 0x004210EE
    PROJECTILE_UPDATE_BLOODY_MESS_DECAL_DX_2 = 0x00421130
    PROJECTILE_UPDATE_BLOODY_MESS_DECAL_DY_2 = 0x0042114B
    PROJECTILE_UPDATE_STOP_ON_HIT_JITTER = 0x004211C4
    PROJECTILE_UPDATE_POST_HIT_DECAL_BURN = 0x004215D2
    PROJECTILE_UPDATE_DEFAULT_FREEZE_SHARD_ANGLE = 0x004215FF
    PROJECTILE_UPDATE_DECAL_SPREAD = 0x00421642
    PROJECTILE_UPDATE_LARGE_STREAK_DIST = 0x0042176F
    PROJECTILE_UPDATE_LARGE_STREAK_DIST_GT4 = 0x00421799
    PROJECTILE_UPDATE_LARGE_STREAK_DIST_GT7 = 0x004217C6
    PROJECTILE_UPDATE_LARGE_STREAK_BURN = 0x0042184C
    PROJECTILE_UPDATE_LARGE_STREAK_FREEZE_ANGLE = 0x00421879
    PROJECTILE_UPDATE_HIT_SFX = 0x00421936
    PROJECTILE_UPDATE_PARTICLE_JITTER_FLAMETHROWER = 0x0042277A
    PROJECTILE_UPDATE_PARTICLE_JITTER_BUBBLEGUN = 0x004227AE
    PROJECTILE_UPDATE_PARTICLE_JITTER_ALT = 0x004227F9
    PROJECTILE_UPDATE_PARTICLE_BOUNCE_SPEED_SCALE = 0x00422A01
    PROJECTILE_UPDATE_PARTICLE_SPRITE_VEL_X = 0x00422BA8
    PROJECTILE_UPDATE_PARTICLE_SPRITE_VEL_Y = 0x00422BBC
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DX_1 = 0x00421EEB
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DY_1 = 0x00421F0A
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DX_2 = 0x00421F47
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DY_2 = 0x00421F66
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DX_3 = 0x00421FAC
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_DECAL_DY_3 = 0x00421FCB
    SECONDARY_PROJECTILE_UPDATE_PRE_HIT_FREEZE_SHARD_ANGLE = 0x00421EB3
    SECONDARY_PROJECTILE_UPDATE_ROCKET_FREEZE_SHARD_ANGLE = 0x00422130
    SECONDARY_PROJECTILE_UPDATE_ROCKET_DECAL_ANGLE = 0x00422165
    SECONDARY_PROJECTILE_UPDATE_ROCKET_DECAL_RADIUS = 0x00422184
    SECONDARY_PROJECTILE_UPDATE_SEEKER_ROCKET_FREEZE_SHARD_ANGLE = 0x0042221E
    SECONDARY_PROJECTILE_UPDATE_SEEKER_ROCKET_DECAL_ANGLE = 0x00422253
    SECONDARY_PROJECTILE_UPDATE_SEEKER_ROCKET_DECAL_RADIUS = 0x00422272
    SECONDARY_PROJECTILE_UPDATE_ROCKET_MINIGUN_FREEZE_SHARD_ANGLE = 0x0042230D
    SECONDARY_PROJECTILE_UPDATE_ROCKET_MINIGUN_DECAL_ANGLE = 0x0042233F
    SECONDARY_PROJECTILE_UPDATE_ROCKET_MINIGUN_DECAL_RADIUS = 0x0042235E
    SECONDARY_PROJECTILE_UPDATE_DETONATION_SPRITE_MAG = 0x004223BE

    UNLOCK_TERRAIN_Q4 = 0x00418229
    UNLOCK_TERRAIN_Q3 = 0x00418254
    UNLOCK_TERRAIN_Q2 = 0x0041827F

    PLAYER_TAKE_DAMAGE_NINJA = 0x00425F23
    PLAYER_TAKE_DAMAGE_DODGER = 0x00425F4E
    PLAYER_TAKE_DAMAGE_HIGHLANDER = 0x00425F79
    PLAYER_TAKE_DAMAGE_DEATH_SFX = 0x00426124
    PLAYER_TAKE_DAMAGE_PAIN_SFX = 0x00426152
    PLAYER_TAKE_DAMAGE_HEADING = 0x00426197
    PLAYER_TAKE_DAMAGE_LOW_HEALTH = 0x00426201

    FX_QUEUE_ADD_RANDOM_GRAY = 0x00427760
    FX_QUEUE_ADD_RANDOM_WIDTH = 0x0042778E
    FX_QUEUE_ADD_RANDOM_ROTATION = 0x004277B0
    FX_QUEUE_ADD_RANDOM_EFFECT_ID = 0x0042780B

    FX_SPAWN_PARTICLE_ALLOC = 0x00420151
    FX_SPAWN_PARTICLE_SPIN = 0x00420206
    FX_SPAWN_PARTICLE_SLOW_ALLOC = 0x00420261
    FX_SPAWN_PARTICLE_SLOW_SPIN = 0x00420316
    FX_SPAWN_SPRITE_ALLOC = 0x0041FBCE
    FX_SPAWN_SPRITE_ROTATION = 0x0041FC3E

    EFFECT_SPAWN_BLOOD_SPLATTER_ROTATION = 0x0042EBC0
    EFFECT_SPAWN_BLOOD_SPLATTER_HALF = 0x0042EBE3
    EFFECT_SPAWN_BLOOD_SPLATTER_SPEED_X = 0x0042EC00
    EFFECT_SPAWN_BLOOD_SPLATTER_SPEED_Y = 0x0042EC1D
    EFFECT_SPAWN_BLOOD_SPLATTER_SCALE_STEP = 0x0042EC44

    EFFECT_SPAWN_BURST_ROTATION = 0x0042EFF2
    EFFECT_SPAWN_BURST_VEL_X = 0x0042F00E
    EFFECT_SPAWN_BURST_VEL_Y = 0x0042F027
    EFFECT_SPAWN_BURST_SCALE_STEP = 0x0042F040

    EFFECT_SPAWN_FREEZE_SHARD_LIFETIME = 0x0042ECD8
    EFFECT_SPAWN_FREEZE_SHARD_ROTATION = 0x0042ED26
    EFFECT_SPAWN_FREEZE_SHARD_HALF = 0x0042ED4B
    EFFECT_SPAWN_FREEZE_SHARD_ROTATION_STEP = 0x0042ED93
    EFFECT_SPAWN_FREEZE_SHARD_SCALE_STEP = 0x0042EDC0
    EFFECT_SPAWN_FREEZE_SHARD_EFFECT_ID = 0x0042EDE3

    EFFECT_SPAWN_FREEZE_SHATTER_HALF = 0x0042EEB8
    EFFECT_SPAWN_FREEZE_SHATTER_ROTATION_STEP = 0x0042EEDC
    EFFECT_SPAWN_FREEZE_SHATTER_SHARD_ANGLE = 0x0042EF27

    EFFECT_SPAWN_EXPLOSION_BURST_PUFF_ROTATION = 0x0042F848
    EFFECT_SPAWN_EXPLOSION_BURST_ROTATION = 0x0042F9E6
    EFFECT_SPAWN_EXPLOSION_BURST_VEL_X = 0x0042FA07
    EFFECT_SPAWN_EXPLOSION_BURST_VEL_Y = 0x0042FA21
    EFFECT_SPAWN_EXPLOSION_BURST_SCALE_STEP = 0x0042FA3B
    EFFECT_SPAWN_EXPLOSION_BURST_ROTATION_STEP = 0x0042FA58

    PERK_SELECT_RANDOM = 0x0042FBDC
    PERKS_GENERATE_CHOICES_RARITY_GATE = 0x004046D4
    PERK_APPLY_FATAL_LOTTERY = 0x0040562F
    PERK_APPLY_BANDAGE_HEAL = 0x004058C9
    PERKS_UPDATE_EFFECTS_REGENERATION_GATE = 0x00406B62
    PERKS_UPDATE_EFFECTS_PYROKINETIC_ANGLE_0P8 = 0x00406DFF
    PERKS_UPDATE_EFFECTS_PYROKINETIC_ANGLE_0P6 = 0x00406E41
    PERKS_UPDATE_EFFECTS_PYROKINETIC_ANGLE_0P4 = 0x00406E83
    PERKS_UPDATE_EFFECTS_PYROKINETIC_ANGLE_0P3 = 0x00406EC5
    PERKS_UPDATE_EFFECTS_PYROKINETIC_ANGLE_0P2 = 0x00406F07
    PERKS_UPDATE_EFFECTS_JINXED_ACCIDENT_GATE = 0x00406FB4
    PERKS_UPDATE_EFFECTS_JINXED_TIMER_RESET = 0x00406FEF
    PERKS_UPDATE_EFFECTS_JINXED_CREATURE_PICK = 0x00407033
    PERKS_UPDATE_EFFECTS_JINXED_CREATURE_RETRY = 0x00407055
    QUEST_START_SELECTED_HIGHSCORE_RANDOM_TAG = 0x0043A81B
    QUEST_BUILD_TARGET_PRACTICE_ANGLE = 0x00437A19
    QUEST_BUILD_TARGET_PRACTICE_RADIUS = 0x00437A38
    QUEST_BUILD_THE_RANDOM_FACTOR_BRUTE_GATE = 0x004363DF
    QUEST_BUILD_SWEEP_STAKES_ANGLE = 0x00437838
    QUEST_BUILD_DEJA_VU_ANGLE = 0x00437946
    QUEST_BUILD_THE_KILLING_TEMPLATE_PICK = 0x004384BE
    QUEST_BUILD_THE_KILLING_LAYOUT_PICK = 0x004384F0
    QUEST_BUILD_THE_KILLING_SPAWNER_1_Y = 0x004385D7
    QUEST_BUILD_THE_KILLING_SPAWNER_1_X = 0x004385F5
    QUEST_BUILD_THE_KILLING_SPAWNER_2_Y = 0x00438628
    QUEST_BUILD_THE_KILLING_SPAWNER_2_X = 0x00438646
    QUEST_BUILD_THE_KILLING_SPAWNER_3_Y = 0x0043867F
    QUEST_BUILD_THE_KILLING_SPAWNER_3_X = 0x0043869D

    TYPO_WORD_PICK_FRAGMENT = 0x00444F88
    TYPO_WORD_PICK_HIGHSCORE_NAME = 0x004452EB
    TYPO_TARGET_NAME_ASSIGN_RANDOM_HIGHSCORE_GATE = 0x004453A8
    TYPO_TARGET_NAME_ASSIGN_RANDOM_FOUR_WORD_GATE = 0x004453F6
    TYPO_TARGET_NAME_ASSIGN_RANDOM_THREE_WORD_GATE_GT80 = 0x0044545B
    TYPO_TARGET_NAME_ASSIGN_RANDOM_THREE_WORD_GATE_GT60 = 0x004454B5
    TYPO_TARGET_NAME_ASSIGN_RANDOM_TWO_WORD_GATE_GT40 = 0x004454D0
    TYPO_TARGET_NAME_ASSIGN_RANDOM_TWO_WORD_GATE_GT20 = 0x0044551C
    TYPO_GAMEPLAY_TYPECLICK_BACKSPACE = 0x0044593F
    TYPO_GAMEPLAY_TYPECLICK_CHAR = 0x004459A1
    CREATURE_SPAWN_TINTED_HEADING = 0x004448A3
    CREATURE_SPAWN_TINTED_SIZE = 0x004448FE
    UI_TEXT_INPUT_UPDATE_TYPECLICK = 0x0043EDC2
    CREDITS_SECRET_ALIEN_ZOOKEEPER_FILL_EMPTY = 0x0040FD61
    CREDITS_SECRET_ALIEN_ZOOKEEPER_REROLL_FILL = 0x0040FE2C
    SFX_PLAY_EXCLUSIVE_PLAYLIST_PICK = 0x0043D4BD

    CAMERA_UPDATE_OFFSET_X_BASE = 0x00409578
    CAMERA_UPDATE_OFFSET_X_SPREAD = 0x004095A7
    CAMERA_UPDATE_OFFSET_X_SIGN = 0x004095BA
    CAMERA_UPDATE_OFFSET_Y_BASE = 0x004095D9
    CAMERA_UPDATE_OFFSET_Y_SPREAD = 0x00409608
    CAMERA_UPDATE_OFFSET_Y_SIGN = 0x0040961B
    DEMO_SETUP_VARIANT_1_SPIDER_SP1_X = 0x0040311C
    DEMO_SETUP_VARIANT_1_SPIDER_SP1_Y = 0x00403138
    DEMO_SETUP_VARIANT_1_SPIDER_SP2_X = 0x0040317E
    DEMO_SETUP_VARIANT_1_SPIDER_SP2_Y = 0x0040319A
    DEMO_SETUP_VARIANT_3_ALIEN_BIG_X = 0x0040327C
    DEMO_SETUP_VARIANT_3_ALIEN_BIG_Y = 0x00403298
    DEMO_SETUP_VARIANT_3_ALIEN_SMALL_X = 0x004032DB
    DEMO_SETUP_VARIANT_3_ALIEN_SMALL_Y = 0x004032F7

    SHRINKIFIER_HIT_ROTATION = 0x0042F1CE
    SHRINKIFIER_HIT_VEL_X = 0x0042F1EA
    SHRINKIFIER_HIT_VEL_Y = 0x0042F209
    SHRINKIFIER_HIT_SCALE_STEP = 0x0042F228

    SPLITTER_HIT_ANGLE = 0x0042F4A2
    SPLITTER_HIT_RADIUS = 0x0042F4C4
    SPLITTER_HIT_AGE = 0x0042F4F3

    ION_HIT_SPARK_ROTATION = 0x0042F61A
    ION_HIT_SPARK_VEL_X = 0x0042F636
    ION_HIT_SPARK_VEL_Y = 0x0042F659
    ION_HIT_SPARK_SCALE_STEP = 0x0042F67C

    WEAPON_PICK_RANDOM_AVAILABLE_PICK = 0x00452CD6
    WEAPON_PICK_RANDOM_AVAILABLE_REROLL_GATE = 0x00452CF1
    WEAPON_PICK_RANDOM_AVAILABLE_REROLL_PICK = 0x00452CFA

    # Synthetic caller ids reserve the high range so they cannot be
    # mistaken for native return addresses from the main executable.
    REWRITE_JINXED_ACCIDENT_TARGET_PICK = 0xF1000001
    REPLAY_DRIVER_INTER_TICK_DRAW_BY_TICK = 0xF1000004
    REPLAY_DRIVER_INTER_TICK_DRAW = 0xF1000005
    REWRITE_STATS_MENU_EASTER_ROLL = 0xF1000006
    REWRITE_STATS_MENU_EASTER_TEXT_X = 0xF1000007


__all__ = ["RngCallerStatic"]



from __future__ import annotations

import msgspec

from grim.config import CrimsonConfig
from grim.geom import Vec2
from grim.raylib_api import rl

from ..frame import RenderFrame
from . import viewport
from .context import build_world_render_ctx
from .draw import draw_world


class WorldRenderer(msgspec.Struct):
    world_size: float
    config: CrimsonConfig | None
    camera: Vec2

    def sync_viewport(
        self,
        *,
        world_size: float,
        config: CrimsonConfig | None,
        camera: Vec2,
    ) -> None:
        self.world_size = float(world_size)
        self.config = config
        self.camera = camera

    def draw(
        self,
        *,
        render_frame: RenderFrame,
        draw_aim_indicators: bool = True,
        entity_alpha: float = 1.0,
    ) -> None:
        self.sync_viewport(
            world_size=render_frame.world_size,
            config=render_frame.config,
            camera=render_frame.camera,
        )
        render_ctx = build_world_render_ctx(self, render_frame=render_frame)
        draw_world(
            render_ctx,
            draw_aim_indicators=draw_aim_indicators,
            entity_alpha=entity_alpha,
        )

    def _camera_screen_size(
        self,
        *,
        runtime_w: float | None = None,
        runtime_h: float | None = None,
    ) -> Vec2:
        out_w = runtime_w if runtime_w is not None else float(rl.get_screen_width())
        out_h = runtime_h if runtime_h is not None else float(rl.get_screen_height())
        return viewport.camera_screen_size(
            world_size=self.world_size,
            config=self.config,
            runtime_w=out_w,
            runtime_h=out_h,
        )

    def _clamp_camera(self, camera: Vec2, screen_size: Vec2) -> Vec2:
        return viewport.clamp_camera(
            world_size=self.world_size,
            camera=camera,
            screen_size=screen_size,
        )

    def _world_params(self) -> tuple[Vec2, Vec2]:
        out_size = Vec2(float(rl.get_screen_width()), float(rl.get_screen_height()))
        camera, view_scale, _screen_size = viewport.view_transform(
            world_size=self.world_size,
            config=self.config,
            camera=self.camera,
            out_size=out_size,
        )
        return camera, view_scale

    def world_to_screen(self, pos: Vec2) -> Vec2:
        camera, view_scale = self._world_params()
        return viewport.world_to_screen_with(pos, camera=camera, view_scale=view_scale)

    def screen_to_world(self, pos: Vec2) -> Vec2:
        camera, view_scale = self._world_params()
        return viewport.screen_to_world_with(pos, camera=camera, view_scale=view_scale)



from __future__ import annotations

from typing import Literal

import msgspec

from grim.assets import TextureId
from grim.audio import AudioState, play_music
from grim.config import (
    CrimsonConfig,
)
from grim.console import ConsoleState
from grim.fonts.grim_mono import GrimMonoFont, load_grim_mono_font
from grim.geom import Vec2
from grim.rand import Crand
from grim.raylib_api import rl
from grim.sfx_map import SfxId
from grim.view import ViewContext

from ..debug import debug_enabled
from ..game_modes import GameMode
from ..net.rollback_resync_v5 import (
    ModeStateSnapshotV2,
    QuestsRuntimeSnapshotV2,
    QuestsStateSnapshotV2,
)
from ..perks.selection import perk_selection_prepared_choices
from ..persistence.save_status import GameStatus, GameStatusData
from ..quests import quest_by_level
from ..quests.level import QuestLevel
from ..quests.runtime import build_quest_spawn_table
from ..quests.status import tracked_quest_games_counter_index
from ..quests.types import QuestContext, QuestDefinition, SpawnEntry
from ..replay import Replay, ReplayHeader, ReplayRecorder
from ..replay.checkpoints import DEFAULT_CHECKPOINT_SAMPLE_RATE
from ..rng_caller_static import RngCallerStatic
from ..sim.bootstrap import advance_explicit_terrain, advance_unlock_terrain
from ..sim.hooks import TickResult
from ..sim.presentation_reactions import (
    PostApplyReaction,
    apply_post_apply_reaction,
    build_post_apply_reaction,
)
from ..sim.session_builders import build_quest_session
from ..sim.sessions import DeterministicSession, DeterministicSessionTick, QuestSpawnState
from ..ui.cursor import draw_menu_cursor
from ..ui.hud import HudRenderContext, draw_hud_overlay, hud_flags_for_game_mode
from ..ui.overlays.quest_run import (
    draw_quest_complete_banner_overlay,
    draw_quest_title_timer_overlay,
)
from ..weapon_runtime import most_used_weapon_id_for_player, weapon_assign_player
from ..weapons import WEAPON_BY_ID, WeaponId
from .base_gameplay_mode import (
    BaseGameplayMode,
    LanSession,
    LanStepAction,
)
from .components.highscore_record_builder import shots_from_state
from .components.perk_menu_controller import PerkMenuController
from .components.perk_prompt_controller import PerkPromptState

WORLD_SIZE = 1024.0

UI_TEXT_SCALE = 1.0
UI_TEXT_COLOR = rl.Color(220, 220, 220, 255)
UI_HINT_COLOR = rl.Color(140, 140, 140, 255)
UI_SPONSOR_COLOR = rl.Color(255, 255, 255, int(255 * 0.5))

_DEBUG_WEAPON_IDS = tuple(sorted(WEAPON_BY_ID))


class QuestRunOutcome(msgspec.Struct, frozen=True):
    kind: str  # "completed" | "failed"
    level: QuestLevel
    base_time_ms: int
    player_health: float
    player2_health: float | None
    pending_perk_count: int
    experience: int
    kill_count: int
    weapon_id: WeaponId
    shots_fired: int
    shots_hit: int
    most_used_weapon_id: WeaponId
    player_health_values: tuple[float, ...] = ()


class QuestMode(BaseGameplayMode):
    def __init__(
        self,
        ctx: ViewContext,
        *,
        demo_mode_active: bool = False,
        config: CrimsonConfig,
        console: ConsoleState | None = None,
        audio: AudioState | None = None,
        audio_rng: Crand,
    ) -> None:
        super().__init__(
            ctx,
            world_size=WORLD_SIZE,
            default_game_mode_id=GameMode.QUESTS,
            demo_mode_active=bool(demo_mode_active),
            quest_fail_retry_count=0,
            hardcore=False,
            config=config,
            console=console,
            audio=audio,
            audio_rng=audio_rng,
        )
        self._quest_def: QuestDefinition | None = None
        self._quest_level: QuestLevel | None = self.config.gameplay.quest_level or QuestLevel(1, 1)
        self._quest_total_spawn_count: int = 0
        self._outcome: QuestRunOutcome | None = None
        self._grim_mono: GrimMonoFont | None = None
        self._perk_prompt = PerkPromptState()
        self._perk_menu = PerkMenuController(on_close=self._reset_perk_prompt)
        self._quest_spawn_state = QuestSpawnState()
        self._sim_session: DeterministicSession | None = None
        self._replay_recorder: ReplayRecorder | None = None

    def open(self) -> None:
        super().open()
        self._quest_def = None
        self._quest_level = self.config.gameplay.quest_level or QuestLevel(1, 1)
        self._quest_total_spawn_count = 0
        self._outcome = None
        self._grim_mono = load_grim_mono_font(self._assets_root)

        self._perk_prompt.reset()
        self._perk_menu.reset()
        self._reset_gameplay_frame_clock()
        self._reset_lan_capture_clock()
        self._replay_recorder = None
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None
        self._sim_session = None

    def close(self) -> None:
        self._grim_mono = None
        self._sim_session = None
        super().close()

    def _new_sim_session(self, *, spawn_entries: tuple[SpawnEntry, ...]) -> DeterministicSession:
        quest_def = self._quest_def
        session, quest_spawn_state = build_quest_session(
            world=self.sim_world.world_state,
            world_size=float(self.world_size),
            damage_scale_by_type=self.sim_world.damage_scale_by_type,
            detail_preset=5,
            violence_disabled=0,
            game_tune_started=bool(self.sim_world.game_tune_started),
            demo_mode_active=bool(self.demo_mode_active),
            apply_world_dt_steps=False,
            finalize_post_render_lifecycle=True,
            spawn_entries=tuple(spawn_entries),
            quest_level=(None if quest_def is None else quest_def.level),
            start_weapon_id=(None if quest_def is None else quest_def.start_weapon_id),
        )
        self._quest_spawn_state = quest_spawn_state
        return session

    def _try_open_perk_menu(self) -> None:
        self._open_perk_menu_ui(
            menu=self._perk_menu,
            players=self.sim_world.players,
            game_mode=GameMode.QUESTS,
            player_count=max(1, len(self.sim_world.players)),
        )

    def _reset_perk_prompt(self) -> None:
        self._perk_prompt.reset_if_pending(pending_count=int(self.state.perk_selection.pending_count))

    def _update_perk_ui(self, *, dt_ui_ms: float) -> None:
        perk_ctx = self._perk_menu_ui_context()
        pending_count = int(self.state.perk_selection.pending_count)
        choices = perk_selection_prepared_choices(self.sim_world.players, self.state.perk_selection)
        self._perk_prompt.begin_frame()
        if self._perk_menu.open:
            choice_index = self._perk_menu.handle_input(
                perk_ctx,
                choices,
                dt_ui_ms=float(dt_ui_ms),
            )
            if choice_index is not None:
                self.record_perk_pick_command(int(choice_index), player_index=0)
        if self._perk_prompt.poll_open_request(
            ctx=perk_ctx,
            config=self.config,
            pending_count=pending_count,
            player_count=max(1, len(self.sim_world.players)),
            any_alive=self._any_player_alive(),
            paused=self._paused,
            menu_active=self._perk_menu.active,
            prompt_scale=UI_TEXT_SCALE,
        ):
            self._try_open_perk_menu()
        self._perk_prompt.tick_timer(
            pending_count=pending_count,
            any_alive=self._any_player_alive(),
            paused=self._paused,
            menu_active=self._perk_menu.active,
            dt_ui_ms=float(dt_ui_ms),
        )
        if not self._paused:
            self._perk_prompt.tick_pulse(float(dt_ui_ms))
        self._perk_menu.tick_timeline(float(dt_ui_ms))

    def _replay_checkpoint_elapsed_ms(self) -> float:
        return float(self._quest_spawn_state.spawn_timeline_ms)

    def _replay_claimed_stats_complete(self) -> bool:
        return bool(self._outcome is not None)

    def _replay_claimed_stats_elapsed_ms(self) -> int:
        return int(self._quest_spawn_state.spawn_timeline_ms)

    def _replay_output_basename(self, *, stamp: str, replay: Replay) -> str:
        replay_level = (
            ""
            if replay.header.quest_level is None
            else replay.header.quest_level.text
        )
        level = self._quest_level.text if self._quest_level is not None else (replay_level or "quest")
        kind = str(self._outcome.kind) if self._outcome is not None else "quest"
        base_time_ms = int(self._quest_spawn_state.spawn_timeline_ms)
        return f"quest_{level}_{stamp}_{kind}_t{base_time_ms}"

    def _replay_skip_save_when_empty(self, *, recorder: ReplayRecorder) -> bool:
        # Avoid emitting empty replays/checkpoint sidecars (usually indicates a
        # test harness calling failure/complete helpers without ticking).
        return int(recorder.tick_index) <= 0

    def _lan_mode_name(self) -> Literal["quests"]:
        return "quests"

    def _lan_match_session(self) -> DeterministicSession | None:
        return self._sim_session

    def _lan_on_paused(self, dt: float) -> None:
        self._tick_death_timers(dt, rate=1.0)
        if self._death_transition_ready():
            self._close_failed_run()

    def _lan_prepare_frame(
        self,
        role: str,
        dt_ui_ms: float,
        session: LanSession,
        dt_tick: float,
    ) -> bool:
        _ = role, dt_ui_ms, dt_tick
        session.detail_preset = int(self._deterministic_detail_preset())
        session.violence_disabled = int(self._deterministic_violence_disabled())
        return True

    def _lan_on_tick_applied(
        self,
        tick: DeterministicSessionTick,
        frame_tick_index: int | None,
        dt_tick: float,
    ) -> LanStepAction:
        session = self._sim_session
        _ = tick
        spawn_state = self._quest_spawn_state
        _ = dt_tick
        if frame_tick_index is not None:
            self._store_net_runtime_snapshot(
                snapshot=QuestsStateSnapshotV2(
                    tick_index=int(frame_tick_index),
                    replay_state=self._net_replay_snapshot_state(),
                    runtime_state=QuestsRuntimeSnapshotV2(
                        elapsed_ms=float(session.elapsed_ms if session is not None else 0.0),
                        spawn_entries=tuple(spawn_state.spawn_entries),
                        spawn_timeline_ms=float(spawn_state.spawn_timeline_ms),
                        no_creatures_timer_ms=float(spawn_state.no_creatures_timer_ms),
                        completion_transition_ms=float(spawn_state.completion_transition_ms),
                        perk_pending_count=int(self.state.perk_selection.pending_count),
                    ),
                ),
            )

        if spawn_state.completed:
            if self._outcome is None:
                assert self._quest_level is not None, "quest outcome requires active quest level"
                fired, hit = shots_from_state(self.state, player_index=int(self.player.index))
                most_used_weapon_id = most_used_weapon_id_for_player(
                    self.state,
                    player_index=int(self.player.index),
                    fallback_weapon_id=self.player.weapon.weapon_id,
                )
                player_health_values = tuple(float(player.health) for player in self.sim_world.players)
                player2_health = None
                if len(player_health_values) >= 2:
                    player2_health = float(player_health_values[1])
                self._outcome = QuestRunOutcome(
                    kind="completed",
                    level=self._quest_level,
                    base_time_ms=int(spawn_state.spawn_timeline_ms),
                    player_health=float(player_health_values[0] if player_health_values else self.player.health),
                    player2_health=player2_health,
                    player_health_values=player_health_values,
                    pending_perk_count=int(self.state.perk_selection.pending_count),
                    experience=int(self.player.experience),
                    kill_count=int(self.creatures.kill_count),
                    weapon_id=self.player.weapon.weapon_id,
                    shots_fired=fired,
                    shots_hit=hit,
                    most_used_weapon_id=most_used_weapon_id,
                )
            self._save_replay()
            self.close_requested = True
            return "stop_after_finalize"

        if self._death_transition_ready():
            self._close_failed_run()
            return "stop_after_finalize"
        return "continue"

    def _build_tick_post_apply_reaction(self, *, tick_result: TickResult) -> PostApplyReaction:
        return build_post_apply_reaction(
            tick_result=tick_result,
            quest_state=self._quest_spawn_state,
        )

    def _apply_tick_post_apply_reaction(self, reaction: PostApplyReaction, *, dt_seconds: float) -> None:
        _ = dt_seconds
        apply_post_apply_reaction(
            reaction=reaction,
            play_sfx=self.audio_bridge.router.play_sfx,
            play_completion_music=self._play_quest_completion_music,
        )

    def _play_quest_completion_music(self) -> None:
        if self.audio is None:
            return
        play_music(self.audio, "crimsonquest")
        playback = self.audio.music.playbacks.get("crimsonquest")
        if playback is None:
            return
        playback.volume = 0.0
        try:
            rl.set_music_volume(playback.music, 0.0)
        except RuntimeError:
            playback.volume = 0.0

    def _apply_resync_snapshot(self, snapshot: ModeStateSnapshotV2) -> None:
        if not isinstance(snapshot, QuestsStateSnapshotV2):
            return
        rs = snapshot.runtime_state
        self._quest_spawn_state.spawn_entries = tuple(rs.spawn_entries)
        self._quest_spawn_state.spawn_timeline_ms = float(rs.spawn_timeline_ms)
        self._quest_spawn_state.no_creatures_timer_ms = float(rs.no_creatures_timer_ms)
        self._quest_spawn_state.completion_transition_ms = float(rs.completion_transition_ms)
        self._quest_spawn_state.completed = False
        self._quest_spawn_state.play_hit_sfx = False
        self._quest_spawn_state.play_completion_music = False
        session = self._sim_session
        if session is not None:
            session.elapsed_ms = float(rs.elapsed_ms)

    def consume_outcome(self) -> QuestRunOutcome | None:
        outcome = self._outcome
        self._outcome = None
        return outcome

    def start_run(self, level: QuestLevel, *, status: GameStatus | None) -> None:
        quest = quest_by_level(level)
        if quest is None:
            self._quest_def = None
            self._quest_level = level
            self._quest_total_spawn_count = 0
            self._sim_session = None
            return
        self._outcome = None
        self._replay_recorder = None
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None

        hardcore_flag = self.config.gameplay.hardcore

        self.hardcore = hardcore_flag
        # Native quest start does not reseed RNG per level; carry the current
        # session RNG state into the next run.
        seed = int(self.state.rng.state) & 0xFFFFFFFF
        self._run_reset_seed = int(seed)

        player_count = self.config.gameplay.player_count
        self._sync_world_runtime_config()
        self.world_runtime.reset(seed=seed, player_count=max(1, min(4, player_count)))
        self._bind_world()
        self._local_input.reset(players=self.sim_world.players)
        self.bind_status(status)
        bound_status = self.state.status
        generic_unlock_index = int(bound_status.quest_unlock_index) if bound_status is not None else 0
        advance_unlock_terrain(
            self.state.rng,
            unlock_index=int(generic_unlock_index),
            width=int(self.world_size),
            height=int(self.world_size),
        )
        # Native `quest_start_selected()` burns one `crt_rand()` for
        # `highscore_record_random_tag` before quest terrain and spawn setup.
        self.state.rng.rand(caller=RngCallerStatic.QUEST_START_SELECTED_HIGHSCORE_RANDOM_TAG)
        quest_terrain = advance_explicit_terrain(
            self.state.rng,
            terrain_slots=quest.terrain_slots,
            width=int(self.world_size),
            height=int(self.world_size),
        )
        self.apply_terrain_setup(
            terrain_slots=quest_terrain.terrain_slots,
            seed=int(quest_terrain.terrain_seed),
        )

        ctx = QuestContext(
            width=int(self.world_size),
            height=int(self.world_size),
            player_count=len(self.sim_world.players),
        )
        entries = build_quest_spawn_table(
            quest,
            ctx,
            rng=self.state.rng,
            hardcore=hardcore_flag,
            full_version=not self.demo_mode_active,
        )
        self.sim_world.state.rng.srand(int(self.state.rng.state))
        total_spawn_count = sum(int(entry.count) for entry in entries)
        self._quest_def = quest
        self._quest_level = quest.level
        self._quest_total_spawn_count = int(total_spawn_count)
        self._reset_gameplay_frame_clock()
        self._sim_session = self._new_sim_session(spawn_entries=tuple(entries))

        replay_status = GameStatusData() if status is None else status.as_data()
        record_replay = (not bool(self._lan_enabled)) or str(self._lan_role) == "host"
        if record_replay:
            self._replay_recorder = ReplayRecorder(
                ReplayHeader(
                    game_mode_id=GameMode.QUESTS,
                    seed=int(self._run_reset_seed),
                    quest_level=quest.level,
                    tick_rate=int(self._gameplay_tick_rate()),
                    quest_fail_retry_count=int(self.quest_fail_retry_count),
                    hardcore=bool(self.hardcore),
                    preserve_bugs=bool(self.state.preserve_bugs),
                    detail_preset=self.config.display.detail_preset,
                    violence_disabled=self.config.display.violence_disabled,
                    world_size=float(self.world_size),
                    player_count=len(self.sim_world.players),
                    status=replay_status,
                ),
            )
            self._replay_checkpoints_sample_rate = int(DEFAULT_CHECKPOINT_SAMPLE_RATE)
        else:
            self._replay_recorder = None
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None

        if status is not None:
            idx = tracked_quest_games_counter_index(quest.level)
            if idx is not None:
                status.increment_quest_play_count(idx)

    def _handle_input(self) -> None:
        if self._perk_menu.open and rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self.audio_bridge.router.play_sfx(SfxId.UI_BUTTONCLICK)
            self._perk_menu.close()
            return

        if (not bool(self._lan_enabled)) and rl.is_key_pressed(rl.KeyboardKey.KEY_TAB):
            self._paused = not self._paused

        if debug_enabled() and (not self._perk_menu.open):
            if rl.is_key_pressed(rl.KeyboardKey.KEY_F2):
                self.state.debug_god_mode = not bool(self.state.debug_god_mode)
                self.audio_bridge.router.play_sfx(SfxId.UI_BUTTONCLICK)
            if rl.is_key_pressed(rl.KeyboardKey.KEY_F3):
                self.state.perk_selection.pending_count += 1
                self.state.perk_selection.choices_dirty = True
                self.audio_bridge.router.play_sfx(SfxId.UI_LEVELUP)
            if rl.is_key_pressed(rl.KeyboardKey.KEY_LEFT_BRACKET):
                self._debug_cycle_weapon(-1)
            if rl.is_key_pressed(rl.KeyboardKey.KEY_RIGHT_BRACKET):
                self._debug_cycle_weapon(1)

        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self._action = "open_pause_menu"
            return

    def _debug_cycle_weapon(self, delta: int) -> None:
        weapon_ids = _DEBUG_WEAPON_IDS
        if not weapon_ids:
            return
        current = self.player.weapon.weapon_id
        try:
            idx = weapon_ids.index(current)
        except ValueError:
            idx = 0
        weapon_id = WeaponId(weapon_ids[(idx + int(delta)) % len(weapon_ids)])
        weapon_assign_player(self.player, weapon_id, state=self.state)

    def _death_transition_ready(self) -> bool:
        dead_players = 0
        for player in self.sim_world.players:
            if float(player.health) > 0.0:
                return False
            dead_players += 1
            if float(player.death_timer) >= 0.0:
                return False
        return dead_players > 0

    def _tick_death_timers(self, dt: float, *, rate: float = 20.0) -> None:
        delta = float(dt) * float(rate)
        if delta <= 0.0:
            return
        for player in self.sim_world.players:
            if float(player.health) > 0.0:
                continue
            if float(player.death_timer) < 0.0:
                continue
            player.death_timer = float(player.death_timer) - delta

    def _close_failed_run(self) -> None:
        if self._outcome is None:
            assert self._quest_level is not None, "quest outcome requires active quest level"
            fired, hit = shots_from_state(self.state, player_index=int(self.player.index))
            most_used_weapon_id = most_used_weapon_id_for_player(
                self.state,
                player_index=int(self.player.index),
                fallback_weapon_id=self.player.weapon.weapon_id,
            )
            player_health_values = tuple(float(player.health) for player in self.sim_world.players)
            player2_health = None
            if len(player_health_values) >= 2:
                player2_health = float(player_health_values[1])
            self._outcome = QuestRunOutcome(
                kind="failed",
                level=self._quest_level,
                base_time_ms=int(self._quest_spawn_state.spawn_timeline_ms),
                player_health=float(player_health_values[0] if player_health_values else self.player.health),
                player2_health=player2_health,
                player_health_values=player_health_values,
                pending_perk_count=int(self.state.perk_selection.pending_count),
                experience=int(self.player.experience),
                kill_count=int(self.creatures.kill_count),
                weapon_id=self.player.weapon.weapon_id,
                shots_fired=fired,
                shots_hit=hit,
                most_used_weapon_id=most_used_weapon_id,
            )
        self._save_replay()
        self.close_requested = True

    def update(self, dt: float) -> None:
        frame = self._begin_mode_update(float(dt))
        if frame is None:
            return
        if bool(self.close_requested):
            return
        if bool(self._lan_enabled) and self._lan_runtime is not None:
            self._update_lan_match(dt=float(frame.dt), dt_ui_ms=float(frame.dt_ui_ms))
            return

        self._update_perk_ui(dt_ui_ms=float(frame.dt_ui_ms))

        sim_dt = 0.0 if (self._paused or self._perk_menu.active) else float(frame.dt)
        session = self._sim_session
        if self._lan_wait_gate_active():
            self._reset_gameplay_frame_clock()
            return
        if sim_dt <= 0.0:
            self._reset_gameplay_frame_clock()
            # Match legacy transition behavior: keep countdown moving, but at
            # real-time pace while perk-menu transition is holding world ticks.
            self._tick_death_timers(float(frame.dt), rate=1.0)
            if self._death_transition_ready():
                self._close_failed_run()
            return
        if session is None:
            self._tick_death_timers(float(sim_dt))
            if self._death_transition_ready():
                self._close_failed_run()
            return

        session.detail_preset = int(self._deterministic_detail_preset())
        session.violence_disabled = int(self._deterministic_violence_disabled())

        self._world_runtime.sync_audio_bridge_state()
        if self.render_resources.ground is not None:
            self.render_resources.ground.process_pending()

        tick_dt = float(self._gameplay_tick_dt(session=session))

        def _on_tick(tick, tick_index: int | None) -> bool:
            _ = tick_index
            action = self._lan_on_tick_applied(tick, None, tick_dt)
            return action != "continue"

        def _on_checkpoint(tick_index: int, tick) -> None:
            self._record_replay_checkpoint_from_tick(
                tick_index=int(tick_index),
                tick=tick,
            )

        self._run_deterministic_session_ticks(
            dt_frame=float(sim_dt),
            session=session,
            recorder=self._replay_recorder,
            on_tick=_on_tick,
            on_checkpoint=_on_checkpoint,
        )

    def draw(self) -> None:
        perk_menu_active = self._perk_menu.active
        debug_overlay_height = 0.0
        self._draw_world(
            draw_aim_indicators=not perk_menu_active,
            entity_alpha=self._world_entity_alpha(),
        )
        self._draw_screen_fade()

        hud_bottom = 0.0
        if not perk_menu_active:
            total = int(self._quest_total_spawn_count)
            kills = int(self.creatures.kill_count)
            quest_progress_ratio = float(kills) / float(total) if total > 0 else None
            hud_flags = hud_flags_for_game_mode(self._config_game_mode_id())
            self._draw_target_health_bar()
            hud_bottom = draw_hud_overlay(
                HudRenderContext(
                    resources=self.render_resources.resources,
                    state=self._hud_state,
                    font=self._small,
                    show_health=hud_flags.show_health,
                    show_weapon=hud_flags.show_weapon,
                    show_xp=hud_flags.show_xp,
                    show_time=hud_flags.show_time,
                    show_quest_hud=hud_flags.show_quest_hud,
                    small_indicators=self._hud_small_indicators(),
                ),
                player=self.player,
                players=self.sim_world.players,
                bonus_hud=self.state.bonus_hud,
                elapsed_ms=float(self._quest_spawn_state.spawn_timeline_ms),
                frame_dt_ms=self._last_dt_ms,
                quest_progress_ratio=quest_progress_ratio,
            )

        if debug_enabled() and (not perk_menu_active):
            x = 18.0
            y = max(18.0, hud_bottom + 10.0)
            god = "on" if self.state.debug_god_mode else "off"
            line = float(self._ui_line_height(scale=0.9))
            self._draw_ui_text(f"debug: [/] weapon  F3 perk+1  F2 god={god}", Vec2(x, y), UI_HINT_COLOR, scale=0.9)
            overlay_end_y = self._draw_lan_debug_info(x=x, y=y + line, line_h=line)
            debug_overlay_height = max(0.0, float(overlay_end_y) - float(y))

        self._draw_quest_title()
        self._draw_quest_complete_banner()

        self._perk_prompt.draw(
            ctx=self._perk_menu_ui_context(),
            pending_count=int(self.state.perk_selection.pending_count),
            any_alive=self._any_player_alive(),
            menu_active=self._perk_menu.active,
            config=self.config,
            ui_text_width=self._ui_text_width,
            text_color=UI_TEXT_COLOR,
            prompt_scale=UI_TEXT_SCALE,
        )
        self._perk_menu.draw(
            self._perk_menu_ui_context(),
            perk_selection_prepared_choices(self.sim_world.players, self.state.perk_selection),
        )

        if perk_menu_active:
            self._draw_game_cursor()
        elif self._paused:
            self._draw_game_cursor()
            x = 18.0
            y = max(18.0, hud_bottom + 10.0)
            y += float(debug_overlay_height)
            self._draw_ui_text("paused (TAB)", Vec2(x, y), UI_HINT_COLOR)
        self._draw_lan_wait_overlay()

    def _draw_game_cursor(self) -> None:
        resources = self.render_resources.resources
        mouse_pos = self._ui_mouse
        draw_menu_cursor(
            resources.texture(TextureId.PARTICLES),
            resources.texture(TextureId.UI_CURSOR),
            pos=mouse_pos,
            pulse_time=float(self._cursor_pulse_time),
        )

    def _draw_quest_title(self) -> None:
        font = self._grim_mono
        quest = self._quest_def
        if font is None or quest is None:
            return
        draw_quest_title_timer_overlay(
            font,
            quest.title,
            quest.level.text,
            timer_ms=float(self._quest_spawn_state.spawn_timeline_ms),
        )

    def _draw_quest_complete_banner(self) -> None:
        tex = self.render_resources.resources.texture(TextureId.UI_TEXT_LEVEL_COMPLETE)
        draw_quest_complete_banner_overlay(
            tex,
            timer_ms=float(self._quest_spawn_state.completion_transition_ms),
        )



from __future__ import annotations

from grim.assets import TextureId
from grim.audio import AudioState
from grim.config import CrimsonConfig
from grim.console import ConsoleState
from grim.rand import Crand
from grim.raylib_api import rl
from grim.view import ViewContext

from ..game_modes import GameMode
from ..persistence.highscores import scores_path_for_mode
from ..persistence.save_status import GameStatusData
from ..replay import Replay, ReplayHeader, ReplayRecorder
from ..replay.checkpoints import DEFAULT_CHECKPOINT_SAMPLE_RATE
from ..sim.bootstrap import advance_unlock_terrain
from ..sim.input import PlayerInput
from ..sim.input_providers import TypoBackspaceCommand, TypoCharCommand, TypoSubmitCommand
from ..sim.session_builders import build_typo_session
from ..sim.sessions import DeterministicSession
from ..typo.names import load_typo_dictionary, load_typo_highscore_names
from ..typo.player import build_typo_player_input
from ..typo.state import typo_shot_counts
from ..ui.cursor import draw_menu_cursor
from ..ui.hud import HudRenderContext, draw_hud_overlay, hud_flags_for_game_mode
from ..ui.overlays.typo_run import draw_typing_box, draw_typo_name_labels
from .base_gameplay_mode import BaseGameplayMode
from .components.highscore_record_builder import build_highscore_record_for_game_over

WORLD_SIZE = 1024.0


class TypoShooterMode(BaseGameplayMode):
    def __init__(
        self,
        ctx: ViewContext,
        *,
        config: CrimsonConfig,
        console: ConsoleState | None = None,
        audio: AudioState | None = None,
        audio_rng: Crand,
    ) -> None:
        super().__init__(
            ctx,
            world_size=WORLD_SIZE,
            default_game_mode_id=GameMode.TYPO,
            demo_mode_active=False,
            quest_fail_retry_count=0,
            hardcore=False,
            config=config,
            console=console,
            audio=audio,
            audio_rng=audio_rng,
        )
        self._sim_session: DeterministicSession | None = self._new_sim_session()

    def _new_sim_session(self) -> DeterministicSession:
        return build_typo_session(
            world=self.sim_world.world_state,
            world_size=float(self.world_size),
            damage_scale_by_type=self.sim_world.damage_scale_by_type,
            detail_preset=5,
            violence_disabled=0,
            game_tune_started=bool(self.sim_world.game_tune_started),
            dictionary_words=self.state.typo.dictionary_words,
            highscore_names=self.state.typo.highscore_names,
        )

    def open(self) -> None:
        super().open()
        dictionary_path = self._base_dir / "typo_dictionary.txt"
        dictionary_words: tuple[str, ...] = ()
        if dictionary_path.is_file():
            dictionary_words = tuple(load_typo_dictionary(dictionary_path))
        highscore_names = tuple(load_typo_highscore_names(scores_path_for_mode(self._base_dir, GameMode.TYPO)))

        status = self.state.status
        terrain = advance_unlock_terrain(
            self.state.rng,
            unlock_index=int(status.quest_unlock_index) if status is not None else 0,
            width=int(self.world_size),
            height=int(self.world_size),
        )
        self.apply_terrain_setup(
            terrain_slots=terrain.terrain_slots,
            seed=int(terrain.terrain_seed),
        )
        self.sim_world.state.rng.srand(int(self.state.rng.state))
        self.state.typo.dictionary_words = dictionary_words
        self.state.typo.highscore_names = highscore_names
        self._sim_session = self._new_sim_session()

        self._replay_recorder = ReplayRecorder(
            ReplayHeader(
                game_mode_id=GameMode.TYPO,
                seed=int(self._run_reset_seed),
                tick_rate=int(self._gameplay_tick_rate()),
                quest_fail_retry_count=int(self.quest_fail_retry_count),
                hardcore=bool(self.hardcore),
                preserve_bugs=bool(self.state.preserve_bugs),
                detail_preset=int(self._deterministic_detail_preset()),
                violence_disabled=int(self._deterministic_violence_disabled()),
                world_size=float(self.world_size),
                player_count=1,
                status=GameStatusData() if status is None else status.as_data(),
                typo_dictionary_words=tuple(dictionary_words),
                typo_highscore_names=tuple(highscore_names),
            ),
        )
        self._replay_checkpoints_sample_rate = int(DEFAULT_CHECKPOINT_SAMPLE_RATE)
        self._replay_checkpoints.clear()
        self._replay_checkpoints_last_tick = None

    def close(self) -> None:
        self._sim_session = None
        super().close()

    def _runtime_player_count(self) -> int:
        return 1

    def _build_local_inputs(self, *, dt: float) -> list[PlayerInput]:
        _ = dt
        return [
            build_typo_player_input(
                aim=self.screen_to_world(self._ui_mouse),
                fire_requested=False,
                reload_requested=False,
            ),
        ]

    def _handle_input(self) -> None:
        if self._game_over_active:
            if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
                self._action = "back_to_menu"
                self.close_requested = True
            return

        if rl.is_key_pressed(rl.KeyboardKey.KEY_TAB):
            self._paused = not self._paused

        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self._action = "open_pause_menu"
            return

    def _enqueue_typing_commands(self) -> None:
        enter_pressed = rl.is_key_pressed(rl.KeyboardKey.KEY_ENTER) or rl.is_key_pressed(rl.KeyboardKey.KEY_KP_ENTER)
        if enter_pressed and self.state.typo.typing.text:
            self.enqueue_input_command(TypoSubmitCommand(player_index=0))

        # Native processes at most one keychar per frame via `console_input_poll`.
        if rl.is_key_pressed(rl.KeyboardKey.KEY_BACKSPACE) or rl.is_key_pressed_repeat(rl.KeyboardKey.KEY_BACKSPACE):
            self.enqueue_input_command(TypoBackspaceCommand(player_index=0))
        else:
            codepoint = int(rl.get_char_pressed())
            if codepoint not in (13, 8) and 0x20 <= codepoint <= 0xFF:
                try:
                    ch = chr(codepoint)
                except ValueError:
                    ch = ""
                if ch:
                    self.enqueue_input_command(TypoCharCommand(player_index=0, ch=ch[0]))

    def _enter_game_over(self) -> None:
        if self._game_over_active:
            return

        shots_fired, shots_hit = typo_shot_counts(self.state.typo)
        record = build_highscore_record_for_game_over(
            state=self.state,
            player=self.player,
            survival_elapsed_ms=int(self._session_elapsed_ms()),
            creature_kill_count=int(self.creatures.kill_count),
            game_mode_id=GameMode.TYPO,
            shots_fired=int(shots_fired),
            shots_hit=int(shots_hit),
            clamp_shots_hit=False,
        )

        self._game_over_record = record
        self._game_over_ui.open()
        self._game_over_active = True
        self._save_replay()

    def _replay_checkpoint_elapsed_ms(self) -> float:
        return self._session_elapsed_ms()

    def _replay_claimed_stats_complete(self) -> bool:
        return bool(self._game_over_active)

    def _replay_claimed_stats_elapsed_ms(self) -> int:
        return int(self._session_elapsed_ms())

    def _replay_claimed_shots(self) -> tuple[int, int]:
        return typo_shot_counts(self.state.typo)

    def _replay_output_basename(self, *, stamp: str, replay: Replay) -> str:
        _ = replay
        score = int(self.player.experience)
        return f"typo_{stamp}_score{score}"

    def update(self, dt: float) -> None:
        self._update_audio(dt)

        dt = self._tick_frame(dt)[0]
        self._handle_input()
        if self._action == "open_pause_menu":
            return

        if self._game_over_active:
            self._update_game_over_ui(dt)
            return

        dt_world = 0.0 if self._paused else dt

        # Native: delay game-over transition until the trooper death animation finishes
        # (checks `death_timer < 0.0` in the main gameplay loop).
        if self.player.health <= 0.0:
            if dt_world > 0.0:
                self.player.death_timer -= float(dt_world) * 20.0
            if self.player.death_timer < 0.0:
                self._enter_game_over()
                self._update_game_over_ui(dt)
                return
            return

        if dt_world > 0.0:
            self._enqueue_typing_commands()

        if dt_world <= 0.0:
            return

        session = self._sim_session
        if session is None:
            return

        def _on_checkpoint(tick_index: int, tick) -> None:
            self._record_replay_checkpoint_from_tick(
                tick_index=int(tick_index),
                tick=tick,
            )

        self._run_deterministic_session_ticks(
            dt_frame=float(dt_world),
            session=session,
            recorder=self._replay_recorder,
            on_tick=lambda _tick, _tick_index: False,
            on_checkpoint=_on_checkpoint,
        )
        # Death/game-over flow is handled at the start of the next frame so the
        # trooper death animation can play before the UI slides in.

    def _draw_game_cursor(self) -> None:
        resources = self.render_resources.resources
        mouse_pos = self._ui_mouse
        draw_menu_cursor(
            resources.texture(TextureId.PARTICLES),
            resources.texture(TextureId.UI_CURSOR),
            pos=mouse_pos,
            pulse_time=float(self._cursor_pulse_time),
        )

    def _draw_name_labels(self) -> None:
        draw_typo_name_labels(
            creatures=self.creatures.entries,
            names=self.state.typo.names.names,
            world_to_screen=self.world_to_screen,
            draw_text=lambda text, pos, color, scale: self._draw_ui_text(text, pos, color, scale=scale),
            measure_text_width=lambda text, scale: float(self._ui_text_width(text, scale=scale)),
        )

    def _draw_typing_box(self) -> None:
        draw_typing_box(
            self.render_resources.resources.texture(TextureId.UI_IND_PANEL),
            text=self.state.typo.typing.text,
            cursor_pulse_time=float(self._cursor_pulse_time),
            draw_text=lambda text, pos, color, scale: self._draw_ui_text(text, pos, color, scale=scale),
            measure_text_width=lambda text, scale: float(self._ui_text_width(text, scale=scale)),
        )

    def draw(self) -> None:
        alive = self.player.health > 0.0
        show_gameplay_ui = alive and (not self._game_over_active)

        self._draw_world(draw_aim_indicators=show_gameplay_ui, entity_alpha=self._world_entity_alpha())
        self._draw_screen_fade()

        if show_gameplay_ui:
            self._draw_name_labels()

        if show_gameplay_ui:
            hud_flags = hud_flags_for_game_mode(self._config_game_mode_id())
            self._draw_target_health_bar()
            draw_hud_overlay(
                HudRenderContext(
                    resources=self.render_resources.resources,
                    state=self._hud_state,
                    font=self._small,
                    show_health=hud_flags.show_health,
                    show_weapon=hud_flags.show_weapon,
                    show_xp=hud_flags.show_xp,
                    show_time=hud_flags.show_time,
                    show_quest_hud=hud_flags.show_quest_hud,
                    small_indicators=self._hud_small_indicators(),
                ),
                player=self.player,
                players=self.sim_world.players,
                bonus_hud=self.state.bonus_hud,
                elapsed_ms=float(self._session_elapsed_ms()),
                frame_dt_ms=self._last_dt_ms,
            )

        if show_gameplay_ui:
            self._draw_typing_box()

        if self._game_over_active:
            self._draw_game_cursor()
            if self._game_over_record is not None:
                self._game_over_ui.draw(
                    record=self._game_over_record,
                    banner_kind=self._game_over_banner,
                    resources=self.render_resources.resources,
                    mouse=self._ui_mouse_pos(),
                )



from __future__ import annotations

import msgspec


class TutorialState(msgspec.Struct):
    stage_index: int = -1
    stage_timer_ms: int = 0
    stage_transition_timer_ms: int = -1000
    hint_index: int = -1
    hint_alpha: int = 0
    hint_fade_in: bool = False
    repeat_spawn_count: int = 0
    hint_bonus_creature_ref: int | None = None
    preserve_bugs: bool = False
    move_active_this_tick: bool = False
    fire_active_this_tick: bool = False
    hint_bonus_alive_before_tick: bool = False


class TutorialOverlayState(msgspec.Struct):
    prompt_text: str = ""
    prompt_alpha: float = 0.0
    hint_text: str = ""
    hint_alpha: float = 0.0


def reset_tutorial_state(
    tutorial: TutorialState,
    overlay: TutorialOverlayState,
    *,
    preserve_bugs: bool,
) -> None:
    tutorial.stage_index = -1
    tutorial.stage_timer_ms = 0
    tutorial.stage_transition_timer_ms = -1000
    tutorial.hint_index = -1
    tutorial.hint_alpha = 0
    tutorial.hint_fade_in = False
    tutorial.repeat_spawn_count = 0
    tutorial.hint_bonus_creature_ref = None
    tutorial.preserve_bugs = bool(preserve_bugs)
    tutorial.move_active_this_tick = False
    tutorial.fire_active_this_tick = False
    tutorial.hint_bonus_alive_before_tick = False
    overlay.prompt_text = ""
    overlay.prompt_alpha = 0.0
    overlay.hint_text = ""
    overlay.hint_alpha = 0.0



from __future__ import annotations

import re
from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Protocol

import msgspec
import typer
from tqdm import tqdm

from ..game_modes import GameMode
from ..paths import default_runtime_dir
from ..quests.level import QuestLevel
from ..weapons import WeaponId

if TYPE_CHECKING:
    from ..dbg.checkpoint_diff import ReplayDiffResult
    from ..replay import Replay
    from ..replay.driver.replay_benchmark import (
        BenchmarkAggregate,
        BenchmarkSample,
        ReplayProfileResult,
        ReplayRenderTelemetryArtifacts,
        ReplayRenderTelemetryFrame,
        ReplayRenderTelemetryTopTick,
    )
    from ..replay.driver.replay_info import ReplayInfoResult, ReplayInfoTimelineEvent
    from ..replay.driver.replay_render import ReplayRenderPhase
    from ..replay.driver.setup import RunResult
    from ..replay.types import ReplayClaimedStatsSnapshot

_REPLAY_VERIFY_SCHEMA_VERSION = 2
_REPLAY_INFO_SCHEMA_VERSION = 2
_REPLAY_BENCHMARK_SCHEMA_VERSION = 3
_REPLAY_VERIFY_SCORE_MISMATCH_EXIT_CODE = 3


class _ProgressBarLike(Protocol):
    total: int

    def update(self, value: int) -> None: ...

    def set_postfix(self, *, refresh: bool = True, **kwargs: object) -> None: ...

    def close(self) -> None: ...


ReplayRenderProgressCallback = Callable[["ReplayRenderPhase", int, int, int], None]
ReplayRenderProgressClose = Callable[[], None]


def _resolve_replay_path(replay_file: Path, *, base_dir: Path) -> tuple[Path, tuple[Path, ...]]:
    """Resolve a replay path, with a convenience lookup under the runtime dir.

    If the input is just a filename and it doesn't exist in the current directory,
    try `base_dir/replays/<name>`.
    """

    path = Path(replay_file)
    tried: list[Path] = [path]
    if path.is_file():
        return path, tuple(tried)

    if not path.is_absolute() and len(path.parts) == 1:
        under_replays = base_dir / "replays" / path.name
        if under_replays not in tried:
            tried.append(under_replays)
            if under_replays.is_file():
                return under_replays, tuple(tried)

    return path, tuple(tried)


def _default_replay_render_output_path(replay_path: Path) -> Path:
    return Path(replay_path).with_suffix(".render.mp4")


def _replay_render_progress_callback(
    *,
    total_ticks: int,
    render_audio: bool,
    tqdm_factory: Callable[..., _ProgressBarLike] = tqdm,
) -> tuple[ReplayRenderProgressCallback | None, ReplayRenderProgressClose | None]:
    if int(total_ticks) <= 0:
        return None, None

    video_bar = tqdm_factory(
        total=int(total_ticks),
        unit="tick",
        desc="replay video",
        leave=True,
    )
    audio_bar: _ProgressBarLike | None = None
    video_last_tick = 0
    audio_last_tick = 0

    def _ensure_audio_bar(total: int) -> _ProgressBarLike:
        nonlocal audio_bar
        if audio_bar is not None:
            return audio_bar
        audio_bar = tqdm_factory(
            total=int(total),
            unit="tick",
            desc="replay audio",
            leave=True,
        )
        return audio_bar

    def callback(phase: str, frame_count: int, tick_index: int, callback_total_ticks: int) -> None:
        nonlocal video_last_tick, audio_last_tick
        resolved_total = int(total_ticks)
        if int(callback_total_ticks) > 0:
            resolved_total = int(callback_total_ticks)
        if int(resolved_total) <= 0:
            return
        if str(phase) == "video":
            bar = video_bar
            last_tick = int(video_last_tick)
        elif str(phase) == "audio":
            if not bool(render_audio):
                return
            bar = _ensure_audio_bar(int(resolved_total))
            last_tick = int(audio_last_tick)
        else:
            return
        if int(bar.total) != int(resolved_total):
            bar.total = int(resolved_total)
        tick = min(int(resolved_total), max(0, int(tick_index)))
        delta = int(tick) - int(last_tick)
        if int(delta) <= 0:
            return
        bar.update(int(delta))
        if str(phase) == "video":
            bar.set_postfix(frames=int(frame_count), refresh=False)
            video_last_tick = int(tick)
        else:
            audio_last_tick = int(tick)

    def close() -> None:
        video_bar.close()
        if audio_bar is not None:
            audio_bar.close()

    return callback, close


def _render_checkpoint_diff_failure(diff: ReplayDiffResult) -> None:
    failure = diff.failure
    assert failure is not None
    exp = failure.expected
    act = failure.actual

    if failure.kind == "missing_checkpoint":
        typer.echo(f"checkpoint missing at tick={int(failure.tick_index)}", err=True)
        raise typer.Exit(code=1)

    assert act is not None
    typer.echo(f"checkpoint mismatch at tick={int(failure.tick_index)}", err=True)
    typer.echo(f"  rng_state expected={exp.rng_state} actual={act.rng_state}", err=True)
    typer.echo(f"  score_xp expected={exp.score_xp} actual={act.score_xp}", err=True)
    typer.echo(f"  kills expected={exp.kills} actual={act.kills}", err=True)
    typer.echo(f"  creature_count expected={exp.creature_count} actual={act.creature_count}", err=True)
    typer.echo(f"  perk_pending expected={exp.perk_pending} actual={act.perk_pending}", err=True)
    typer.echo(f"  deaths expected={len(exp.deaths)} actual={len(act.deaths)}", err=True)
    if exp.deaths or act.deaths:
        typer.echo(f"  first death expected={exp.deaths[:1]} actual={act.deaths[:1]}", err=True)
    if int(exp.events.hit_count) >= 0:
        typer.echo(
            "  events "
            f"expected=(hits={exp.events.hit_count}, pickups={exp.events.pickup_count}, sfx={exp.events.sfx_count}, head={exp.events.sfx_head}) "
            f"actual=(hits={act.events.hit_count}, pickups={act.events.pickup_count}, sfx={act.events.sfx_count}, head={act.events.sfx_head})",
            err=True,
        )
    if exp.perk != act.perk:
        typer.echo(
            "  perk snapshot differs "
            f"(expected pending={exp.perk.pending_count} choices={exp.perk.choices}, "
            f"actual pending={act.perk.pending_count} choices={act.perk.choices})",
            err=True,
        )
    raise typer.Exit(code=1)


def _replay_mode_label(game_mode_id: GameMode) -> str:
    match game_mode_id:
        case GameMode.SURVIVAL:
            return "survival"
        case GameMode.RUSH:
            return "rush"
        case GameMode.QUESTS:
            return "quests"
        case _:
            return "unknown"


class _RunResultPayload(msgspec.Struct, forbid_unknown_fields=True):
    game_mode_id: GameMode
    tick_rate: int
    ticks: int
    elapsed_ms: int
    score_xp: int
    creature_kill_count: int
    most_used_weapon_id: WeaponId
    shots_fired: int
    shots_hit: int
    rng_state: int


class _ReplayVerifyScoreClaimPayload(msgspec.Struct, forbid_unknown_fields=True):
    metric: str
    submitted_score: int
    simulated_value: int
    match: bool


class _ReplayVerifyClaimedStatsPayload(msgspec.Struct, forbid_unknown_fields=True):
    complete: bool
    ticks: int
    elapsed_ms: int
    score_xp: int
    kills: int
    most_used_weapon_id: WeaponId
    shots_fired: int
    shots_hit: int


class _ReplayVerifyHeaderClaimPayload(msgspec.Struct, forbid_unknown_fields=True):
    expected: _ReplayVerifyClaimedStatsPayload
    simulated: _ReplayVerifyClaimedStatsPayload
    match: bool
    mismatched_fields: list[str]


class _ReplayVerifyPayload(msgspec.Struct, forbid_unknown_fields=True):
    schema_version: int
    status: str
    replay: str
    run_result: _RunResultPayload
    header_claim: _ReplayVerifyHeaderClaimPayload | None
    score_claim: _ReplayVerifyScoreClaimPayload | None


class _ReplayInfoSummaryPayload(msgspec.Struct, forbid_unknown_fields=True):
    game_mode_id: GameMode
    tick_rate: int
    ticks_simulated: int
    elapsed_ms: int
    player_count: int
    event_count: int
    event_counts_by_kind: dict[str, int]


class _ReplayInfoEventPayload(msgspec.Struct, forbid_unknown_fields=True):
    tick_index: int
    elapsed_ms: int
    elapsed_s: float
    kind: str
    player_index: int | None
    detail: str
    data: dict[str, object]


class _ReplayInfoPayload(msgspec.Struct, forbid_unknown_fields=True):
    schema_version: int
    status: str
    replay: str
    summary: _ReplayInfoSummaryPayload
    timeline: list[_ReplayInfoEventPayload]


class _BenchmarkAggregatePayload(msgspec.Struct, forbid_unknown_fields=True):
    min: float
    p50: float
    mean: float
    p95: float
    max: float
    stdev: float


class _ReplayBenchmarkProfileHotspotPayload(msgspec.Struct, forbid_unknown_fields=True):
    file: str
    line: int
    function: str
    primitive_calls: int
    total_calls: int
    tottime: float
    cumtime: float


class _ReplayBenchmarkProfilePayload(msgspec.Struct, forbid_unknown_fields=True):
    sort: str
    top: int
    source: str
    hotspots: list[_ReplayBenchmarkProfileHotspotPayload]


class _ReplayBenchmarkSettingsPayload(msgspec.Struct, forbid_unknown_fields=True):
    mode: str
    runs: int
    warmup_runs: int
    max_ticks: int | None
    trace_rng: bool
    profile: bool
    profile_sort: str
    top: int
    profile_out: str | None
    render_telemetry: bool
    render_telemetry_out: str | None
    render_charts_out_dir: str | None


class _ReplayBenchmarkSamplePayload(msgspec.Struct, forbid_unknown_fields=True):
    wall_ms: float
    ticks_per_second: float
    realtime_x: float


class _ReplayBenchmarkSummaryPayload(msgspec.Struct, forbid_unknown_fields=True):
    sample_count: int
    samples: list[_ReplayBenchmarkSamplePayload]
    wall_ms: _BenchmarkAggregatePayload
    ticks_per_second: _BenchmarkAggregatePayload
    realtime_x: _BenchmarkAggregatePayload


class _ReplayRenderTelemetryTopTickPayload(msgspec.Struct, forbid_unknown_fields=True):
    tick_index: int
    frame_index: int
    value: float


class _ReplayRenderTelemetryFramePayload(msgspec.Struct, forbid_unknown_fields=True):
    frame_index: int
    tick_index_before_update: int
    tick_index_after_update: int
    update_ms: float
    draw_ms: float
    frame_ms: float
    draw_calls_total: int
    draw_calls_by_api: dict[str, int]
    draw_calls_by_pass: dict[str, int]
    pass_ms: dict[str, float]


class _ReplayRenderTelemetrySummaryPayload(msgspec.Struct, forbid_unknown_fields=True):
    frame_ms: _BenchmarkAggregatePayload
    update_ms: _BenchmarkAggregatePayload
    draw_ms: _BenchmarkAggregatePayload
    draw_calls_total: _BenchmarkAggregatePayload
    top_draw_ms_ticks: list[_ReplayRenderTelemetryTopTickPayload]
    top_frame_ms_ticks: list[_ReplayRenderTelemetryTopTickPayload]
    top_draw_calls_ticks: list[_ReplayRenderTelemetryTopTickPayload]


class _ReplayRenderTelemetryArtifactsPayload(msgspec.Struct, forbid_unknown_fields=True):
    telemetry_json_path: str | None
    charts_dir: str | None
    frame_timing_svg: str | None
    draw_calls_svg: str | None
    pass_timing_stacked_svg: str | None
    report_md: str | None


class _ReplayRenderTelemetryPayload(msgspec.Struct, forbid_unknown_fields=True):
    summary: _ReplayRenderTelemetrySummaryPayload
    frames: list[_ReplayRenderTelemetryFramePayload]
    preview: list[_ReplayRenderTelemetryFramePayload]
    artifacts: _ReplayRenderTelemetryArtifactsPayload | None


class _ReplayBenchmarkPayload(msgspec.Struct, forbid_unknown_fields=True):
    schema_version: int
    status: str
    replay: str
    settings: _ReplayBenchmarkSettingsPayload
    run_result: _RunResultPayload
    benchmark: _ReplayBenchmarkSummaryPayload
    profile: _ReplayBenchmarkProfilePayload | None
    render_telemetry: _ReplayRenderTelemetryPayload | None


def _run_result_payload(run_result: RunResult) -> _RunResultPayload:
    return _RunResultPayload(
        game_mode_id=run_result.game_mode_id,
        tick_rate=run_result.tick_rate,
        ticks=run_result.ticks,
        elapsed_ms=run_result.elapsed_ms,
        score_xp=run_result.score_xp,
        creature_kill_count=run_result.creature_kill_count,
        most_used_weapon_id=run_result.most_used_weapon_id,
        shots_fired=run_result.shots_fired,
        shots_hit=run_result.shots_hit,
        rng_state=run_result.rng_state,
    )


def _replay_info_event_payload(event: ReplayInfoTimelineEvent) -> _ReplayInfoEventPayload:
    return _ReplayInfoEventPayload(
        tick_index=event.tick_index,
        elapsed_ms=event.elapsed_ms,
        elapsed_s=event.elapsed_ms / 1000.0,
        kind=str(event.kind),
        player_index=event.player_index,
        detail=event.detail,
        data=event.data,
    )


def _replay_verify_claimed_stats_payload(
    claimed_stats: ReplayClaimedStatsSnapshot,
) -> _ReplayVerifyClaimedStatsPayload:
    return _ReplayVerifyClaimedStatsPayload(
        complete=claimed_stats.complete,
        ticks=claimed_stats.ticks,
        elapsed_ms=claimed_stats.elapsed_ms,
        score_xp=claimed_stats.score_xp,
        kills=claimed_stats.kills,
        most_used_weapon_id=claimed_stats.most_used_weapon_id,
        shots_fired=claimed_stats.shots_fired,
        shots_hit=claimed_stats.shots_hit,
    )


def _replay_verify_run_result_claim_payload(
    run_result: RunResult,
    *,
    complete: bool,
) -> _ReplayVerifyClaimedStatsPayload:
    return _ReplayVerifyClaimedStatsPayload(
        complete=complete,
        ticks=run_result.ticks,
        elapsed_ms=run_result.elapsed_ms,
        score_xp=run_result.score_xp,
        kills=run_result.creature_kill_count,
        most_used_weapon_id=run_result.most_used_weapon_id,
        shots_fired=run_result.shots_fired,
        shots_hit=run_result.shots_hit,
    )


def _replay_verify_mismatched_fields(
    expected_claim: _ReplayVerifyClaimedStatsPayload,
    simulated_claim: _ReplayVerifyClaimedStatsPayload,
) -> list[str]:
    mismatched_fields: list[str] = []
    if expected_claim.ticks != simulated_claim.ticks:
        mismatched_fields.append("ticks")
    if expected_claim.elapsed_ms != simulated_claim.elapsed_ms:
        mismatched_fields.append("elapsed_ms")
    if expected_claim.score_xp != simulated_claim.score_xp:
        mismatched_fields.append("score_xp")
    if expected_claim.kills != simulated_claim.kills:
        mismatched_fields.append("kills")
    if expected_claim.most_used_weapon_id != simulated_claim.most_used_weapon_id:
        mismatched_fields.append("most_used_weapon_id")
    if expected_claim.shots_fired != simulated_claim.shots_fired:
        mismatched_fields.append("shots_fired")
    if expected_claim.shots_hit != simulated_claim.shots_hit:
        mismatched_fields.append("shots_hit")
    return mismatched_fields


def _replay_info_summary_payload(
    result: ReplayInfoResult,
    *,
    event_count: int,
    event_counts_by_kind: dict[str, int],
) -> _ReplayInfoSummaryPayload:
    return _ReplayInfoSummaryPayload(
        game_mode_id=result.game_mode_id,
        tick_rate=result.tick_rate,
        ticks_simulated=result.ticks_simulated,
        elapsed_ms=result.elapsed_ms,
        player_count=result.player_count,
        event_count=event_count,
        event_counts_by_kind=event_counts_by_kind,
    )


def _benchmark_aggregate_payload(aggregate: BenchmarkAggregate) -> _BenchmarkAggregatePayload:
    return _BenchmarkAggregatePayload(
        min=aggregate.min,
        p50=aggregate.p50,
        mean=aggregate.mean,
        p95=aggregate.p95,
        max=aggregate.max,
        stdev=aggregate.stdev,
    )


def _render_telemetry_top_tick_payload(entry: ReplayRenderTelemetryTopTick) -> _ReplayRenderTelemetryTopTickPayload:
    return _ReplayRenderTelemetryTopTickPayload(
        tick_index=entry.tick_index,
        frame_index=entry.frame_index,
        value=entry.value,
    )


def _render_telemetry_frame_payload(entry: ReplayRenderTelemetryFrame) -> _ReplayRenderTelemetryFramePayload:
    return _ReplayRenderTelemetryFramePayload(
        frame_index=entry.frame_index,
        tick_index_before_update=entry.tick_index_before_update,
        tick_index_after_update=entry.tick_index_after_update,
        update_ms=entry.update_ms,
        draw_ms=entry.draw_ms,
        frame_ms=entry.frame_ms,
        draw_calls_total=entry.draw_calls_total,
        draw_calls_by_api=dict(entry.draw_calls_by_api),
        draw_calls_by_pass=dict(entry.draw_calls_by_pass),
        pass_ms=dict(entry.pass_ms),
    )


def _path_text(path: Path | None) -> str | None:
    if path is None:
        return None
    return str(path)


def _replay_benchmark_profile_payload(profile: ReplayProfileResult | None) -> _ReplayBenchmarkProfilePayload | None:
    if profile is None:
        return None
    return _ReplayBenchmarkProfilePayload(
        sort=str(profile.sort),
        top=profile.top,
        source=str(profile.source),
        hotspots=[
            _ReplayBenchmarkProfileHotspotPayload(
                file=row.file,
                line=row.line,
                function=row.function,
                primitive_calls=row.primitive_calls,
                total_calls=row.total_calls,
                tottime=row.tottime,
                cumtime=row.cumtime,
            )
            for row in profile.hotspots
        ],
    )


def _replay_benchmark_settings_payload(
    *,
    mode: Literal["headless", "render"],
    runs: int,
    warmup_runs: int,
    max_ticks: int | None,
    trace_rng: bool,
    profile: bool,
    profile_sort: Literal["cumtime", "tottime"],
    top: int,
    profile_out: Path | None,
    render_telemetry: bool,
    render_telemetry_out: Path | None,
    render_charts_out_dir: Path | None,
) -> _ReplayBenchmarkSettingsPayload:
    return _ReplayBenchmarkSettingsPayload(
        mode=mode,
        runs=runs,
        warmup_runs=warmup_runs,
        max_ticks=max_ticks,
        trace_rng=trace_rng,
        profile=profile,
        profile_sort=profile_sort,
        top=top,
        profile_out=_path_text(profile_out),
        render_telemetry=render_telemetry,
        render_telemetry_out=_path_text(render_telemetry_out),
        render_charts_out_dir=_path_text(render_charts_out_dir),
    )


def _replay_benchmark_sample_payload(sample: BenchmarkSample) -> _ReplayBenchmarkSamplePayload:
    return _ReplayBenchmarkSamplePayload(
        wall_ms=sample.wall_ms,
        ticks_per_second=sample.ticks_per_second,
        realtime_x=sample.realtime_x,
    )


def _render_telemetry_artifacts_payload(
    artifacts: ReplayRenderTelemetryArtifacts | None,
) -> _ReplayRenderTelemetryArtifactsPayload | None:
    if artifacts is None:
        return None
    return _ReplayRenderTelemetryArtifactsPayload(
        telemetry_json_path=_path_text(artifacts.telemetry_json_path),
        charts_dir=_path_text(artifacts.charts_dir),
        frame_timing_svg=_path_text(artifacts.frame_timing_svg),
        draw_calls_svg=_path_text(artifacts.draw_calls_svg),
        pass_timing_stacked_svg=_path_text(artifacts.pass_timing_stacked_svg),
        report_md=_path_text(artifacts.report_md),
    )


def _fmt_metric_agg(name: str, aggregate: BenchmarkAggregate, *, digits: int) -> str:
    entry = aggregate
    return (
        f"{name} "
        f"min={float(entry.min):.{digits}f} "
        f"p50={float(entry.p50):.{digits}f} "
        f"mean={float(entry.mean):.{digits}f} "
        f"p95={float(entry.p95):.{digits}f} "
        f"max={float(entry.max):.{digits}f} "
        f"stdev={float(entry.stdev):.{digits}f}"
    )


@dataclass(frozen=True, slots=True)
class _ReplayListRow:
    replay: str
    mode: str
    game_mode_id: GameMode | None
    game_version: str
    ticks: str
    duration: str
    score_xp: str
    kills: str
    modified: str
    modified_ts: float
    old_version: bool


def _fmt_replay_list_duration(*, ticks: int, tick_rate: int) -> str:
    if int(tick_rate) <= 0:
        return "n/a"
    total_seconds = float(ticks) / float(tick_rate)
    if total_seconds >= 3600.0:
        hours = int(total_seconds // 3600.0)
        minutes = int((total_seconds % 3600.0) // 60.0)
        seconds = int(total_seconds % 60.0)
        return f"{hours:d}:{minutes:02d}:{seconds:02d}"
    if total_seconds >= 60.0:
        minutes = int(total_seconds // 60.0)
        seconds = int(total_seconds % 60.0)
        return f"{minutes:d}:{seconds:02d}"
    return f"{total_seconds:.1f}s"


def _fmt_replay_list_modified(timestamp: float) -> str:
    return datetime.fromtimestamp(float(timestamp)).strftime("%Y-%m-%d %H:%M")


def _version_tuple(value: str) -> tuple[int, ...]:
    return tuple(int(part) for part in re.findall(r"\d+", str(value)))


def _is_version_older(*, replay_version: str, current_version: str) -> bool:
    replay_parts = _version_tuple(replay_version)
    current_parts = _version_tuple(current_version)
    if not replay_parts or not current_parts:
        return False
    width = max(len(replay_parts), len(current_parts))
    replay_norm = replay_parts + (0,) * (width - len(replay_parts))
    current_norm = current_parts + (0,) * (width - len(current_parts))
    return replay_norm < current_norm


def _replay_list_mode_label(
    *,
    game_mode_id: GameMode,
    player_count: int,
    quest_level: QuestLevel | None,
) -> str:
    match game_mode_id:
        case GameMode.QUESTS:
            label = "quest"
            if quest_level is not None:
                label = f"{label} {quest_level.text}"
        case _:
            label = _replay_mode_label(game_mode_id)
    if int(player_count) > 1:
        label = f"{label} {int(player_count)}p"
    return label


def _replay_list_mode_style(game_mode_id: GameMode | None) -> str:
    match game_mode_id:
        case GameMode.SURVIVAL:
            return "green"
        case GameMode.RUSH:
            return "magenta"
        case GameMode.QUESTS:
            return "cyan"
        case GameMode.TYPO:
            return "blue"
        case GameMode.TUTORIAL:
            return "white"
        case _:
            return "white"


def _replay_list_score_kills(
    *,
    replay: Replay,
) -> tuple[str, str]:
    claimed_stats = replay.header.claimed_stats
    return str(int(claimed_stats.score_xp)), str(int(claimed_stats.kills))


def _build_replay_list_row(
    replay_path: Path,
    *,
    replays_dir: Path,
    load_replay_fn: Callable[[bytes], Replay],
    current_version: str,
) -> tuple[_ReplayListRow, str | None]:
    rel = str(replay_path.relative_to(replays_dir))
    modified_text = "?"
    modified_ts = 0.0
    try:
        stat = replay_path.stat()
        modified_ts = float(stat.st_mtime)
        modified_text = _fmt_replay_list_modified(modified_ts)
    except OSError as exc:
        return (
            _ReplayListRow(
                replay=rel,
                mode="error",
                game_mode_id=None,
                game_version="-",
                ticks="-",
                duration="-",
                score_xp="-",
                kills="-",
                modified=modified_text,
                modified_ts=float(modified_ts),
                old_version=False,
            ),
            str(exc).replace("\n", " ").strip(),
        )

    try:
        replay = load_replay_fn(replay_path.read_bytes())
    except (OSError, ValueError) as exc:
        return (
            _ReplayListRow(
                replay=rel,
                mode="invalid",
                game_mode_id=None,
                game_version="-",
                ticks="-",
                duration="-",
                score_xp="-",
                kills="-",
                modified=modified_text,
                modified_ts=float(modified_ts),
                old_version=False,
            ),
            str(exc).replace("\n", " ").strip(),
        )

    header = replay.header
    game_mode_id = header.game_mode_id
    tick_rate = int(header.tick_rate)
    ticks = len(replay.ticks)
    game_version = str(header.game_version).strip() or "-"
    player_count = int(header.player_count)
    mode_label = _replay_list_mode_label(
        game_mode_id=game_mode_id,
        player_count=player_count,
        quest_level=header.quest_level,
    )
    score_xp, kills = _replay_list_score_kills(
        replay=replay,
    )
    is_old = _is_version_older(replay_version=game_version, current_version=current_version)
    return (
        _ReplayListRow(
            replay=rel,
            mode=mode_label,
            game_mode_id=game_mode_id,
            game_version=game_version,
            ticks=str(ticks),
            duration=_fmt_replay_list_duration(ticks=ticks, tick_rate=tick_rate),
            score_xp=score_xp,
            kills=kills,
            modified=modified_text,
            modified_ts=float(modified_ts),
            old_version=bool(is_old),
        ),
        None,
    )


replay_app = typer.Typer(add_completion=False)


@replay_app.command("play")
def cmd_replay_play(
    replay_file: Path = typer.Argument(
        ...,
        help="replay file path (.crd); if a filename is provided, also search base-dir/replays",
    ),
    width: int | None = typer.Option(None, help="window width (default: use crimson.cfg)"),
    height: int | None = typer.Option(None, help="window height (default: use crimson.cfg)"),
    fps: int = typer.Option(60, help="target fps"),
    base_dir: Path = typer.Option(
        default_runtime_dir(),
        "--base-dir",
        "--runtime-dir",
        help="base path for runtime files (default: per-user OS data dir; override with CRIMSON_RUNTIME_DIR)",
    ),
    assets_dir: Path | None = typer.Option(
        None,
        help="assets root (default: base-dir; missing .paq files are downloaded)",
    ),
) -> None:
    """Play back a recorded replay."""
    from grim.app import RunViewHooks, run_view
    from grim.config import ensure_crimson_cfg
    from grim.console import create_console
    from grim.view import ViewContext

    from ..assets_fetch import download_missing_paqs
    from ..modes.replay_playback_mode import ReplayPlaybackMode
    from ..runtime_resources_view import RuntimeResourcesView

    if assets_dir is None:
        assets_dir = base_dir
    base_dir.mkdir(parents=True, exist_ok=True)
    replay_path, tried = _resolve_replay_path(replay_file, base_dir=base_dir)
    if not replay_path.is_file():
        message = f"replay file not found: {tried[0]}"
        if len(tried) > 1:
            message += f" (also tried: {tried[1]})"
        typer.echo(message, err=True)
        raise typer.Exit(code=1)
    cfg = ensure_crimson_cfg(base_dir)
    if width is None:
        width = cfg.display.width
    if height is None:
        height = cfg.display.height
    console = create_console(base_dir, assets_dir=assets_dir)
    download_missing_paqs(assets_dir, console)

    ctx = ViewContext(assets_dir=assets_dir, preserve_bugs=False)
    view = ReplayPlaybackMode(ctx, replay_path=replay_path, config=cfg, console=console)
    title = f"Replay — {replay_path.name}"

    run_view(
        RuntimeResourcesView(view, assets_dir=assets_dir),
        width=width,
        height=height,
        title=title,
        fps=fps,
        hooks=RunViewHooks(
            should_close=view.should_close,
            consume_screenshot_request=view.consume_screenshot_request,
        ),
    )


@replay_app.command("list")
def cmd_replay_list(
    base_dir: Path = typer.Option(
        default_runtime_dir(),
        "--base-dir",
        "--runtime-dir",
        help="base path for runtime files (default: per-user OS data dir; override with CRIMSON_RUNTIME_DIR)",
    ),
    color: bool = typer.Option(
        True,
        "--color/--no-color",
        help="enable ANSI colors in table output (default: on)",
    ),
) -> None:
    """List replay files under base-dir/replays."""
    from rich import box
    from rich.console import Console
    from rich.table import Table

    from .. import __version__
    from ..replay import load_replay

    replays_dir = Path(base_dir) / "replays"
    replay_files = sorted(
        (path for path in replays_dir.rglob("*.crd") if path.is_file()),
        key=lambda path: str(path.relative_to(replays_dir)),
    )
    if not replay_files:
        typer.echo(f"no replay files found under {replays_dir}")
        return

    rows: list[_ReplayListRow] = []
    parse_errors: list[str] = []
    for replay_path in replay_files:
        row, parse_error = _build_replay_list_row(
            replay_path,
            replays_dir=replays_dir,
            load_replay_fn=load_replay,
            current_version=str(__version__),
        )
        rows.append(row)
        if parse_error:
            parse_errors.append(f"{row.replay}: {parse_error}")

    rows.sort(key=lambda row: (-float(row.modified_ts), str(row.replay)))

    table = Table(box=box.SIMPLE, header_style="bold")
    table.add_column("replay")
    table.add_column("mode")
    table.add_column("version")
    table.add_column("ticks", justify="right")
    table.add_column("duration", justify="right")
    table.add_column("score", justify="right")
    table.add_column("kills", justify="right")
    table.add_column("modified", style="dim")
    for row in rows:
        mode_style = _replay_list_mode_style(row.game_mode_id)
        mode_cell = f"[{mode_style}]{row.mode}[/{mode_style}]"
        version_cell = row.game_version
        if row.mode in {"invalid", "error"}:
            mode_cell = f"[red]{row.mode}[/red]"
            version_cell = f"[red]{version_cell}[/red]"
        elif bool(row.old_version):
            version_cell = f"[yellow]{version_cell}[/yellow]"
        else:
            version_cell = f"[green]{version_cell}[/green]"
        table.add_row(
            row.replay,
            mode_cell,
            version_cell,
            row.ticks,
            row.duration,
            row.score_xp,
            row.kills,
            row.modified,
            style="dim" if bool(row.old_version) else "",
        )

    console = Console(force_terminal=bool(color), no_color=not bool(color), width=200)
    console.print(table)
    parsed_count = len(rows) - len(parse_errors)
    typer.echo(f"count={len(rows)} parsed={parsed_count} errors={len(parse_errors)}")
    typer.echo(f"replays_dir={replays_dir}")
    for parse_error in parse_errors:
        typer.echo(f"warning: {parse_error}")


@replay_app.command("verify")
def cmd_replay_verify(
    replay_file: Path = typer.Argument(
        ...,
        help="replay file path (.crd); if a filename is provided, also search base-dir/replays",
    ),
    max_ticks: int | None = typer.Option(None, help="stop after N ticks (default: full replay)"),
    trace_rng: bool = typer.Option(
        False,
        "--trace-rng",
        help="enable replay RNG trace mode during simulation",
    ),
    output_format: Literal["human", "json"] = typer.Option(
        "human",
        "--format",
        help="output format",
    ),
    json_out: Path | None = typer.Option(
        None,
        "--json-out",
        help="optional JSON output path for verify result payload",
    ),
    base_dir: Path = typer.Option(
        default_runtime_dir(),
        "--base-dir",
        "--runtime-dir",
        help="base path for runtime files (default: per-user OS data dir; override with CRIMSON_RUNTIME_DIR)",
    ),
) -> None:
    """Headlessly simulate a replay and report resulting run stats."""
    from ..replay import ReplayCodecError, ReplayGameVersionError, load_replay
    from ..replay.driver.playback_driver import build_verify_playback_driver
    from ..replay.driver.setup import ReplayRunnerError

    replay_path, tried = _resolve_replay_path(replay_file, base_dir=base_dir)
    if not replay_path.is_file():
        message = f"replay file not found: {tried[0]}"
        if len(tried) > 1:
            message += f" (also tried: {tried[1]})"
        typer.echo(message, err=True)
        raise typer.Exit(code=1)

    replay_bytes = Path(replay_path).read_bytes()
    try:
        replay = load_replay(replay_bytes)
        result = build_verify_playback_driver(
            replay,
            max_ticks=max_ticks,
            trace_rng=trace_rng,
        ).run()
    except (ReplayCodecError, ReplayGameVersionError, ReplayRunnerError) as exc:
        typer.echo(f"replay verification failed: {exc}", err=True)
        raise typer.Exit(code=1) from exc

    full_replay_simulated = max_ticks is None or max_ticks >= len(replay.ticks)
    header_claim_payload: _ReplayVerifyHeaderClaimPayload | None = None
    header_claim_matches = True
    claimed_stats = replay.header.claimed_stats
    if full_replay_simulated:
        expected_claim = _replay_verify_claimed_stats_payload(claimed_stats)
        simulated_claim = _replay_verify_run_result_claim_payload(result, complete=claimed_stats.complete)
        mismatched_fields = _replay_verify_mismatched_fields(expected_claim, simulated_claim)
        header_claim_matches = not mismatched_fields
        header_claim_payload = _ReplayVerifyHeaderClaimPayload(
            expected=expected_claim,
            simulated=simulated_claim,
            match=header_claim_matches,
            mismatched_fields=mismatched_fields,
        )

    status = "ok" if header_claim_matches else "header_stats_mismatch"

    payload = _ReplayVerifyPayload(
        schema_version=_REPLAY_VERIFY_SCHEMA_VERSION,
        status=status,
        replay=str(replay_path),
        run_result=_run_result_payload(result),
        header_claim=header_claim_payload,
        score_claim=None,
    )
    payload_json = msgspec.json.encode(payload)

    if json_out is not None:
        json_out.parent.mkdir(parents=True, exist_ok=True)
        json_out.write_bytes(payload_json)
        if output_format == "human":
            typer.echo(f"json_report={json_out}")

    if output_format == "json":
        typer.echo(payload_json.decode("utf-8"))
    else:
        message = (
            f"{status}: "
            f"ticks={result.ticks} elapsed_ms={result.elapsed_ms} score_xp={result.score_xp} "
            f"kills={result.creature_kill_count} most_used_weapon_id={result.most_used_weapon_id} "
            f"shots_fired={result.shots_fired} shots_hit={result.shots_hit} rng_state={result.rng_state}"
        )
        if header_claim_payload is not None:
            mismatch_fields = (
                ",".join(header_claim_payload.mismatched_fields) if header_claim_payload.mismatched_fields else "-"
            )
            message += (
                f"; header_claim complete={header_claim_payload.expected.complete} "
                f"match={header_claim_payload.match} "
                f"mismatches={mismatch_fields}"
        )
        typer.echo(message)

    if not header_claim_matches:
        raise typer.Exit(code=_REPLAY_VERIFY_SCORE_MISMATCH_EXIT_CODE)


@replay_app.command("info")
def cmd_replay_info(
    replay_file: Path = typer.Argument(
        ...,
        help="replay file path (.crd); if a filename is provided, also search base-dir/replays",
    ),
    output_format: Literal["human", "json"] = typer.Option(
        "human",
        "--format",
        help="output format",
    ),
    json_out: Path | None = typer.Option(
        None,
        "--json-out",
        help="optional JSON output path for replay info payload",
    ),
    max_ticks: int | None = typer.Option(None, help="stop after N ticks (default: full replay)"),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        help="include extra context events in addition to core gameplay events",
    ),
    player_index: int | None = typer.Option(
        None,
        "--player-index",
        help="optional player index filter (applies to player-specific events)",
    ),
    base_dir: Path = typer.Option(
        default_runtime_dir(),
        "--base-dir",
        "--runtime-dir",
        help="base path for runtime files (default: per-user OS data dir; override with CRIMSON_RUNTIME_DIR)",
    ),
) -> None:
    """Simulate a replay and emit a timeline of gameplay events."""
    from ..replay import ReplayCodecError, ReplayGameVersionError, load_replay
    from ..replay.driver.playback_driver import build_verify_playback_driver
    from ..replay.driver.replay_info import collect_replay_info, event_counts_by_kind
    from ..replay.driver.setup import ReplayRunnerError

    replay_path, tried = _resolve_replay_path(replay_file, base_dir=base_dir)
    if not replay_path.is_file():
        message = f"replay file not found: {tried[0]}"
        if len(tried) > 1:
            message += f" (also tried: {tried[1]})"
        typer.echo(message, err=True)
        raise typer.Exit(code=1)

    replay_bytes = Path(replay_path).read_bytes()
    try:
        replay = load_replay(replay_bytes)
        result = collect_replay_info(
            build_verify_playback_driver(
                replay,
                max_ticks=max_ticks,
                warn_on_version_mismatch=True,
                trace_rng=False,
            ),
            player_index=player_index,
            include_extra_events=bool(verbose),
        )
    except (ReplayCodecError, ReplayGameVersionError, ReplayRunnerError) as exc:
        typer.echo(f"replay info failed: {exc}", err=True)
        raise typer.Exit(code=1) from exc

    timeline_payload = [_replay_info_event_payload(event) for event in result.timeline]
    summary_payload = _replay_info_summary_payload(
        result,
        event_count=len(timeline_payload),
        event_counts_by_kind=event_counts_by_kind(result.timeline),
    )
    payload = _ReplayInfoPayload(
        schema_version=_REPLAY_INFO_SCHEMA_VERSION,
        status="ok",
        replay=str(replay_path),
        summary=summary_payload,
        timeline=timeline_payload,
    )
    payload_json = msgspec.json.encode(payload)

    if json_out is not None:
        json_out.parent.mkdir(parents=True, exist_ok=True)
        json_out.write_bytes(payload_json)

    if output_format == "json":
        typer.echo(payload_json.decode("utf-8"))
        return

    typer.echo(
        "ok: "
        f"replay={replay_path} "
        f"mode={_replay_mode_label(summary_payload.game_mode_id)} "
        f"ticks={summary_payload.ticks_simulated} "
        f"elapsed_ms={summary_payload.elapsed_ms} "
        f"events={summary_payload.event_count}",
    )
    for event in timeline_payload:
        player_tag = f" [p{event.player_index}]" if event.player_index is not None else ""
        typer.echo(
            f"t={event.elapsed_s:.3f} tick={event.tick_index}{player_tag} {event.kind} {event.detail}",
        )

    tail = f"events={summary_payload.event_count}"
    if json_out is not None:
        tail += f" json_report={json_out}"
    typer.echo(tail)


@replay_app.command("benchmark")
def cmd_replay_benchmark(
    replay_file: Path = typer.Argument(
        ...,
        help="replay file path (.crd); if a filename is provided, also search base-dir/replays",
    ),
    runs: int | None = typer.Option(
        None,
        "--runs",
        min=1,
        help="number of measured benchmark runs (default: headless=5, render=1)",
    ),
    warmup_runs: int | None = typer.Option(
        None,
        "--warmup-runs",
        min=0,
        help="warmup runs before measured timing (default: headless=1, render=0)",
    ),
    mode: Literal["headless", "render"] = typer.Option(
        "headless",
        "--mode",
        help="benchmark mode: headless|render",
    ),
    rtx: bool = typer.Option(
        False,
        "--rtx",
        help="enable non-canonical RTX render mode (render mode only)",
    ),
    max_ticks: int | None = typer.Option(None, help="stop after N ticks (default: full replay)"),
    trace_rng: bool = typer.Option(
        False,
        "--trace-rng",
        help="enable replay RNG trace mode during simulation",
    ),
    profile: bool = typer.Option(False, "--profile", help="run one cProfile pass and include hotspot summary"),
    profile_sort: Literal["cumtime", "tottime"] = typer.Option(
        "cumtime",
        "--profile-sort",
        help="hotspot sort key",
    ),
    top: int = typer.Option(20, "--top", min=1, help="maximum hotspot rows to include"),
    profile_out: Path | None = typer.Option(
        None,
        "--profile-out",
        help="optional cProfile .pstats output path (used only with --profile)",
    ),
    render_telemetry: bool = typer.Option(
        False,
        "--render-telemetry",
        help="collect per-tick render telemetry (render mode only)",
    ),
    render_telemetry_out: Path | None = typer.Option(
        None,
        "--render-telemetry-out",
        help="optional output path for full render telemetry JSON (render mode only)",
    ),
    render_charts_out_dir: Path | None = typer.Option(
        None,
        "--render-charts-out-dir",
        help="optional output directory for render telemetry SVG charts (render mode only)",
    ),
    output_format: Literal["human", "json"] = typer.Option(
        "human",
        "--format",
        help="output format",
    ),
    json_out: Path | None = typer.Option(
        None,
        "--json-out",
        help="optional JSON output path for benchmark payload",
    ),
    base_dir: Path = typer.Option(
        default_runtime_dir(),
        "--base-dir",
        "--runtime-dir",
        help="base path for runtime files (default: per-user OS data dir; override with CRIMSON_RUNTIME_DIR)",
    ),
) -> None:
    """Benchmark replay throughput, with optional profiler hotspots."""
    from ..replay import ReplayCodecError, ReplayGameVersionError, load_replay
    from ..replay.driver.replay_benchmark import (
        ReplayBenchmarkError,
        run_replay_benchmark,
        run_replay_render_benchmark,
    )
    from ..replay.driver.setup import ReplayRunnerError

    replay_path, tried = _resolve_replay_path(replay_file, base_dir=base_dir)
    if not replay_path.is_file():
        message = f"replay file not found: {tried[0]}"
        if len(tried) > 1:
            message += f" (also tried: {tried[1]})"
        typer.echo(message, err=True)
        raise typer.Exit(code=1)

    replay_bytes = Path(replay_path).read_bytes()
    resolved_runs = runs if runs is not None else (1 if mode == "render" else 5)
    resolved_warmup_runs = warmup_runs if warmup_runs is not None else (0 if mode == "render" else 1)
    if mode != "render":
        if rtx:
            typer.echo("replay benchmark failed: --rtx is supported only with --mode render", err=True)
            raise typer.Exit(code=1)
        if render_telemetry:
            typer.echo("replay benchmark failed: --render-telemetry is supported only with --mode render", err=True)
            raise typer.Exit(code=1)
        if render_telemetry_out is not None:
            typer.echo("replay benchmark failed: --render-telemetry-out is supported only with --mode render", err=True)
            raise typer.Exit(code=1)
        if render_charts_out_dir is not None:
            typer.echo(
                "replay benchmark failed: --render-charts-out-dir is supported only with --mode render", err=True,
            )
            raise typer.Exit(code=1)

    try:
        replay = load_replay(replay_bytes)
        if mode == "render":
            benchmark = run_replay_render_benchmark(
                replay,
                replay_path=Path(replay_path),
                base_dir=Path(base_dir),
                runs=resolved_runs,
                warmup_runs=resolved_warmup_runs,
                max_ticks=max_ticks,
                trace_rng=trace_rng,
                profile=profile,
                profile_sort=profile_sort,
                top=top,
                profile_out=profile_out,
                render_telemetry=render_telemetry,
                render_telemetry_out=render_telemetry_out,
                render_charts_out_dir=render_charts_out_dir,
                rtx=rtx,
                show_progress=(output_format == "human"),
            )
        else:
            benchmark = run_replay_benchmark(
                replay,
                runs=resolved_runs,
                warmup_runs=resolved_warmup_runs,
                max_ticks=max_ticks,
                trace_rng=trace_rng,
                profile=profile,
                profile_sort=profile_sort,
                top=top,
                profile_out=profile_out,
                show_progress=(output_format == "human"),
            )
    except (ReplayCodecError, ReplayGameVersionError, ReplayBenchmarkError, ReplayRunnerError) as exc:
        typer.echo(f"replay benchmark failed: {exc}", err=True)
        raise typer.Exit(code=1) from exc

    profile_payload = _replay_benchmark_profile_payload(benchmark.profile)

    render_telemetry_payload: _ReplayRenderTelemetryPayload | None = None
    if benchmark.render_telemetry is not None:
        telemetry_summary = benchmark.render_telemetry.summary
        render_telemetry_payload = _ReplayRenderTelemetryPayload(
            summary=_ReplayRenderTelemetrySummaryPayload(
                frame_ms=_benchmark_aggregate_payload(telemetry_summary.frame_ms),
                update_ms=_benchmark_aggregate_payload(telemetry_summary.update_ms),
                draw_ms=_benchmark_aggregate_payload(telemetry_summary.draw_ms),
                draw_calls_total=_benchmark_aggregate_payload(telemetry_summary.draw_calls_total),
                top_draw_ms_ticks=[
                    _render_telemetry_top_tick_payload(entry) for entry in telemetry_summary.top_draw_ms_ticks
                ],
                top_frame_ms_ticks=[
                    _render_telemetry_top_tick_payload(entry) for entry in telemetry_summary.top_frame_ms_ticks
                ],
                top_draw_calls_ticks=[
                    _render_telemetry_top_tick_payload(entry) for entry in telemetry_summary.top_draw_calls_ticks
                ],
            ),
            frames=[_render_telemetry_frame_payload(entry) for entry in benchmark.render_telemetry.frames],
            preview=[_render_telemetry_frame_payload(entry) for entry in benchmark.render_telemetry.preview],
            artifacts=_render_telemetry_artifacts_payload(benchmark.render_telemetry.artifacts),
        )

    payload = _ReplayBenchmarkPayload(
        schema_version=_REPLAY_BENCHMARK_SCHEMA_VERSION,
        status="ok",
        replay=str(replay_path),
        settings=_replay_benchmark_settings_payload(
            mode=mode,
            runs=resolved_runs,
            warmup_runs=resolved_warmup_runs,
            max_ticks=max_ticks,
            trace_rng=trace_rng,
            profile=profile,
            profile_sort=profile_sort,
            top=top,
            profile_out=profile_out,
            render_telemetry=render_telemetry,
            render_telemetry_out=render_telemetry_out,
            render_charts_out_dir=render_charts_out_dir,
        ),
        run_result=_run_result_payload(benchmark.run_result),
        benchmark=_ReplayBenchmarkSummaryPayload(
            sample_count=len(benchmark.samples),
            samples=[_replay_benchmark_sample_payload(sample) for sample in benchmark.samples],
            wall_ms=_benchmark_aggregate_payload(benchmark.wall_ms),
            ticks_per_second=_benchmark_aggregate_payload(benchmark.ticks_per_second),
            realtime_x=_benchmark_aggregate_payload(benchmark.realtime_x),
        ),
        profile=profile_payload,
        render_telemetry=render_telemetry_payload,
    )
    payload_json = msgspec.json.encode(payload)

    if json_out is not None:
        json_out.parent.mkdir(parents=True, exist_ok=True)
        json_out.write_bytes(payload_json)
    if output_format == "human":
            typer.echo(f"json_report={json_out}")

    if output_format == "json":
        typer.echo(payload_json.decode("utf-8"))
        return

    typer.echo(
        "ok: "
        f"mode={mode} "
        f"runs={len(benchmark.samples)} warmup_runs={resolved_warmup_runs} "
        f"ticks={benchmark.run_result.ticks} "
        f"wall_ms_p50={benchmark.wall_ms.p50:.3f} "
        f"tps_p50={benchmark.ticks_per_second.p50:.2f} "
        f"realtime_x_p50={benchmark.realtime_x.p50:.2f}",
    )
    typer.echo(_fmt_metric_agg("wall_ms", benchmark.wall_ms, digits=3))
    typer.echo(
        _fmt_metric_agg("throughput_tps", benchmark.ticks_per_second, digits=2)
        + " | "
        + _fmt_metric_agg("realtime_x", benchmark.realtime_x, digits=2),
    )
    if benchmark.render_telemetry is not None:
        telemetry = benchmark.render_telemetry
        typer.echo(
            "render_telemetry: "
            + _fmt_metric_agg("frame_ms", telemetry.summary.frame_ms, digits=3)
            + " | "
            + _fmt_metric_agg("update_ms", telemetry.summary.update_ms, digits=3)
            + " | "
            + _fmt_metric_agg("draw_ms", telemetry.summary.draw_ms, digits=3),
        )
        typer.echo(_fmt_metric_agg("draw_calls_total", telemetry.summary.draw_calls_total, digits=2))
        typer.echo("render_telemetry top_draw_ms_ticks:")
        if not telemetry.summary.top_draw_ms_ticks:
            typer.echo("  (none)")
        for row in telemetry.summary.top_draw_ms_ticks:
            typer.echo(f"  tick={int(row.tick_index)} frame={int(row.frame_index)} draw_ms={float(row.value):.3f}")
        artifacts = telemetry.artifacts
        if artifacts is not None:
            typer.echo("render_telemetry artifacts:")
            if artifacts.telemetry_json_path:
                typer.echo(f"  telemetry_json={artifacts.telemetry_json_path}")
            if artifacts.charts_dir:
                typer.echo(f"  charts_dir={artifacts.charts_dir}")
            if artifacts.frame_timing_svg:
                typer.echo(f"  frame_timing_svg={artifacts.frame_timing_svg}")
            if artifacts.draw_calls_svg:
                typer.echo(f"  draw_calls_svg={artifacts.draw_calls_svg}")
            if artifacts.pass_timing_stacked_svg:
                typer.echo(f"  pass_timing_stacked_svg={artifacts.pass_timing_stacked_svg}")
            if artifacts.report_md:
                typer.echo(f"  report_md={artifacts.report_md}")
    if benchmark.profile is None:
        return
    typer.echo(
        f"profile: sort={benchmark.profile.sort} source={benchmark.profile.source} top={benchmark.profile.top}",
    )
    typer.echo("hotspots:")
    if not benchmark.profile.hotspots:
        typer.echo("  (none)")
        return
    for idx, row in enumerate(benchmark.profile.hotspots, start=1):
        typer.echo(
            f"  {idx:02d} cum={float(row.cumtime):.6f}s tot={float(row.tottime):.6f}s "
            f"calls={int(row.primitive_calls)}/{int(row.total_calls)} "
            f"{row.file}:{int(row.line)}::{row.function}",
        )


@replay_app.command("render")
def cmd_replay_render(
    replay_file: Path = typer.Argument(
        ...,
        help="replay file path (.crd); if a filename is provided, also search base-dir/replays",
    ),
    out: Path | None = typer.Option(
        None,
        "--out",
        "-o",
        help="output video path (default: <replay>.render.mp4)",
    ),
    width: int | None = typer.Option(None, help="render width (default: use crimson.cfg)"),
    height: int | None = typer.Option(None, help="render height (default: use crimson.cfg)"),
    fps: int = typer.Option(60, "--fps", min=1, help="output video fps"),
    max_ticks: int | None = typer.Option(None, help="stop after N ticks (default: full replay)"),
    trace_rng: bool = typer.Option(
        False,
        "--trace-rng",
        help="enable replay RNG trace mode during simulation",
    ),
    ffmpeg_bin: Path | None = typer.Option(
        None,
        "--ffmpeg-bin",
        help="ffmpeg executable path (default: discover from PATH)",
    ),
    crf: int = typer.Option(
        16,
        "--crf",
        min=0,
        max=51,
        help="ffmpeg quality factor (libx264: lower is higher quality)",
    ),
    preset: Literal[
        "ultrafast",
        "superfast",
        "veryfast",
        "faster",
        "fast",
        "medium",
        "slow",
        "slower",
        "veryslow",
    ] = typer.Option("slow", "--preset", help="ffmpeg libx264 preset"),
    pixel_format: str = typer.Option(
        "yuv420p",
        "--pixel-format",
        help="ffmpeg output pixel format",
    ),
    overwrite: bool = typer.Option(
        False,
        "--overwrite",
        help="overwrite output if it already exists",
    ),
    audio: bool = typer.Option(
        True,
        "--audio/--mute-audio",
        help="include in-game audio in output video",
    ),
    base_dir: Path = typer.Option(
        default_runtime_dir(),
        "--base-dir",
        "--runtime-dir",
        help="base path for runtime files (default: per-user OS data dir; override with CRIMSON_RUNTIME_DIR)",
    ),
    assets_dir: Path | None = typer.Option(
        None,
        help="assets root (default: base-dir; missing .paq files are downloaded)",
    ),
) -> None:
    """Render replay playback to video using ffmpeg."""
    from ..replay import ReplayCodecError, ReplayGameVersionError, load_replay
    from ..replay.driver.replay_render import ReplayRenderError, run_replay_render_video
    from ..replay.driver.setup import ReplayRunnerError

    replay_path, tried = _resolve_replay_path(replay_file, base_dir=base_dir)
    if not replay_path.is_file():
        message = f"replay file not found: {tried[0]}"
        if len(tried) > 1:
            message += f" (also tried: {tried[1]})"
        typer.echo(message, err=True)
        raise typer.Exit(code=1)

    output_path = Path(out) if out is not None else _default_replay_render_output_path(replay_path)

    replay_bytes = Path(replay_path).read_bytes()
    progress_close: ReplayRenderProgressClose | None = None
    try:
        replay = load_replay(replay_bytes)
        total_ticks = len(replay.ticks)
        if max_ticks is not None:
            total_ticks = min(int(total_ticks), max(0, int(max_ticks)))
        progress_callback, progress_close = _replay_render_progress_callback(
            total_ticks=total_ticks,
            render_audio=bool(audio),
        )
        render = run_replay_render_video(
            replay,
            replay_path=Path(replay_path),
            output_path=Path(output_path),
            base_dir=Path(base_dir),
            assets_dir=(Path(assets_dir) if assets_dir is not None else None),
            width=width,
            height=height,
            fps=int(fps),
            max_ticks=max_ticks,
            trace_rng=bool(trace_rng),
            ffmpeg_bin=(Path(ffmpeg_bin) if ffmpeg_bin is not None else None),
            crf=int(crf),
            preset=preset,
            pixel_format=str(pixel_format),
            overwrite=bool(overwrite),
            mute_audio=not bool(audio),
            progress=progress_callback,
        )
    except (ReplayCodecError, ReplayGameVersionError, ReplayRenderError, ReplayRunnerError) as exc:
        typer.echo(f"replay render failed: {exc}", err=True)
        raise typer.Exit(code=1) from exc
    finally:
        if progress_close is not None:
            progress_close()

    message = (
        f"ok: output={render.output_path} "
        f"frames={render.frame_count} fps={render.fps} "
        f"resolution={render.width}x{render.height} "
        f"ticks={render.run_result.ticks} elapsed_ms={render.run_result.elapsed_ms} "
        f"score_xp={render.run_result.score_xp} kills={render.run_result.creature_kill_count}"
    )
    typer.echo(message)


@replay_app.command("verify-checkpoints")
def cmd_replay_verify_checkpoints(
    replay_file: Path = typer.Argument(
        ...,
        help="replay file path (.crd); if a filename is provided, also search base-dir/replays",
    ),
    checkpoints_file: Path | None = typer.Option(
        None,
        "--checkpoints",
        help="checkpoint sidecar path (default: <replay>.chk)",
    ),
    max_ticks: int | None = typer.Option(None, help="stop after N ticks (default: full replay)"),
    trace_rng: bool = typer.Option(
        False,
        "--trace-rng",
        help="include presentation RNG draw marks in verification checkpoints",
    ),
    base_dir: Path = typer.Option(
        default_runtime_dir(),
        "--base-dir",
        "--runtime-dir",
        help="base path for runtime files (default: per-user OS data dir; override with CRIMSON_RUNTIME_DIR)",
    ),
) -> None:
    """Verify a replay by comparing headless checkpoints with a sidecar file."""
    from ..dbg.checkpoint_diff import compare_checkpoints
    from ..replay import ReplayCodecError, ReplayGameVersionError, load_replay
    from ..replay.checkpoints import (
        ReplayCheckpointsError,
        default_checkpoints_path,
        load_checkpoints_file,
    )
    from ..replay.driver.playback_driver import PlaybackWalkHooks, build_verify_playback_driver
    from ..replay.driver.setup import ReplayRunnerError

    replay_path, tried = _resolve_replay_path(replay_file, base_dir=base_dir)
    if not replay_path.is_file():
        message = f"replay file not found: {tried[0]}"
        if len(tried) > 1:
            message += f" (also tried: {tried[1]})"
        typer.echo(message, err=True)
        raise typer.Exit(code=1)

    replay_bytes = Path(replay_path).read_bytes()
    try:
        replay = load_replay(replay_bytes)
    except ReplayCodecError as exc:
        typer.echo(f"replay verification failed: {exc}", err=True)
        raise typer.Exit(code=1) from exc

    if checkpoints_file is None:
        checkpoints_path = default_checkpoints_path(replay_path)
    else:
        checkpoints_path = Path(checkpoints_file)
    if not checkpoints_path.is_file():
        typer.echo(f"checkpoints file not found: {checkpoints_path}", err=True)
        raise typer.Exit(code=1)

    try:
        expected = load_checkpoints_file(checkpoints_path)
    except ReplayCheckpointsError as exc:
        typer.echo(f"replay verification failed: {exc}", err=True)
        raise typer.Exit(code=1) from exc

    checkpoint_ticks = {int(ckpt.tick_index) for ckpt in expected.checkpoints}
    actual = []

    try:
        driver = build_verify_playback_driver(
            replay,
            max_ticks=max_ticks,
            trace_rng=bool(trace_rng),
        )

        def _after_tick(tick_result, _world) -> None:
            tick_index = int(tick_result.source_tick.tick_index)
            if tick_index in checkpoint_ticks:
                actual.append(driver.build_checkpoint(tick_result=tick_result))

        result = driver.run(
            hooks=PlaybackWalkHooks(after_tick=_after_tick),
        )
    except (ReplayGameVersionError, ReplayRunnerError) as exc:
        typer.echo(f"replay verification failed: {exc}", err=True)
        raise typer.Exit(code=1) from exc

    diff = compare_checkpoints(expected.checkpoints, actual)
    if not diff.ok:
        _render_checkpoint_diff_failure(diff)

    message = (
        f"ok: {len(expected.checkpoints)} checkpoints match; ticks={result.ticks} "
        f"score_xp={result.score_xp} kills={result.creature_kill_count}"
    )
    if diff.first_rng_only_tick is not None:
        message += f"; rng-only drift starts at tick={diff.first_rng_only_tick}"
    typer.echo(message)


@replay_app.command("diff-checkpoints")
def cmd_replay_diff_checkpoints(
    expected_file: Path = typer.Argument(..., help="expected checkpoints sidecar (.crd.chk)"),
    actual_file: Path = typer.Argument(..., help="actual checkpoints sidecar (.crd.chk)"),
) -> None:
    """Compare two checkpoint sidecars and report the first divergence."""
    from ..dbg.checkpoint_diff import compare_checkpoints
    from ..replay.checkpoints import load_checkpoints_file

    expected = load_checkpoints_file(Path(expected_file))
    actual = load_checkpoints_file(Path(actual_file))
    diff = compare_checkpoints(expected.checkpoints, actual.checkpoints)
    if not diff.ok:
        _render_checkpoint_diff_failure(diff)

    message = f"ok: {len(expected.checkpoints)} checkpoints match"
    if diff.first_rng_only_tick is not None:
        message += f"; rng-only drift starts at tick={diff.first_rng_only_tick}"
    typer.echo(message)



from __future__ import annotations

from collections.abc import Callable

from grim.config import CrimsonConfig
from grim.rand import CrandLike
from grim.raylib_api import rl
from grim.sfx_map import SfxId

from ..input_codes import INPUT_CODE_UNBOUND, input_code_is_down
from ..rng_caller_static import RngCallerStatic

_CONTROL_BIND_SLOTS = 5
_SINGLE_PLAYER_ALT_MOVE_CODES: tuple[int, ...] = (0xC8, 0xD0, 0xCB, 0xCD)


def poll_text_input(max_len: int, *, allow_space: bool = True) -> str:
    out = ""
    while True:
        value = rl.get_char_pressed()
        if value == 0:
            break
        if value < 0x20 or value > 0xFF:
            continue
        if not allow_space and value == 0x20:
            continue
        if len(out) >= max_len:
            continue
        out += chr(int(value))
    return out


def flush_text_input_events() -> None:
    # Native flows call `grim_flush_input()` before entering high-score name input.
    while rl.get_char_pressed():
        pass
    while rl.get_key_pressed():
        pass


def update_name_entry_text(
    text: str,
    caret: int,
    *,
    max_len: int,
    rng: CrandLike,
    play_sfx: Callable[[SfxId], None] | None = None,
) -> tuple[str, int]:
    typed = poll_text_input(max_len - len(text), allow_space=True)
    if typed:
        text = (text[:caret] + typed + text[caret:])[:max_len]
        caret = min(len(text), caret + len(typed))
        if play_sfx is not None:
            play_sfx(
                SfxId.UI_TYPECLICK_01
                if (rng.rand(caller=RngCallerStatic.UI_TEXT_INPUT_UPDATE_TYPECLICK) & 1) == 0
                else SfxId.UI_TYPECLICK_02,
            )
    if rl.is_key_pressed(rl.KeyboardKey.KEY_BACKSPACE) and caret > 0:
        text = text[: caret - 1] + text[caret:]
        caret -= 1
        if play_sfx is not None:
            play_sfx(
                SfxId.UI_TYPECLICK_01
                if (rng.rand(caller=RngCallerStatic.UI_TEXT_INPUT_UPDATE_TYPECLICK) & 1) == 0
                else SfxId.UI_TYPECLICK_02,
            )
    if rl.is_key_pressed(rl.KeyboardKey.KEY_LEFT):
        caret = max(0, caret - 1)
    if rl.is_key_pressed(rl.KeyboardKey.KEY_RIGHT):
        caret = min(len(text), caret + 1)
    if rl.is_key_pressed(rl.KeyboardKey.KEY_HOME):
        caret = 0
    if rl.is_key_pressed(rl.KeyboardKey.KEY_END):
        caret = len(text)
    return text, caret


def gameplay_controls_held(config: CrimsonConfig) -> bool:
    player_count = max(1, min(4, config.gameplay.player_count))
    for player_index in range(player_count):
        player_controls = config.controls.player(player_index)
        move_forward_key, move_backward_key, turn_left_key, turn_right_key = player_controls.move_codes
        for code in (
            move_forward_key,
            move_backward_key,
            turn_left_key,
            turn_right_key,
            player_controls.fire_code,
        )[:_CONTROL_BIND_SLOTS]:
            if code == INPUT_CODE_UNBOUND:
                continue
            if input_code_is_down(code, player_index=player_index):
                return True

    for code in _SINGLE_PLAYER_ALT_MOVE_CODES:
        if input_code_is_down(int(code), player_index=0):
            return True
    return False



from __future__ import annotations

from pathlib import Path

from .channel_compare import compare_entity_samples, compare_rng_stream, compare_sim_state
from .channel_helpers import (
    ENTITY_SAMPLE_KINDS,
    checkpoint_channel_required,
    entity_rows,
    entity_samples_channel_required,
    rng_stream_channel_required,
    sim_state_channel_required,
)
from .checkpoint_diff import checkpoint_deepdiff
from .payloads import BuiltinObject, to_builtin_object
from .schema import TickRecord
from .trace import TraceReader


def _entity_uid_set(row: TickRecord, kind: str) -> set[int]:
    samples = entity_samples_channel_required(row)
    out: set[int] = set()
    for item in entity_rows(samples, kind=kind):
        out.add(int(item.uid))
    return out


def focus_tick(
    *,
    golden_trace: Path,
    candidate_trace: Path,
    tick_index: int,
) -> BuiltinObject:
    tick = int(tick_index)
    with TraceReader(Path(golden_trace)) as expected, TraceReader(Path(candidate_trace)) as candidate:
        expected_row = expected.tick(tick)
        candidate_row = candidate.tick(tick)

    if expected_row is None or candidate_row is None:
        raise ValueError(f"tick {tick} missing in one of the traces")

    expected_checkpoint = checkpoint_channel_required(expected_row)
    candidate_checkpoint = checkpoint_channel_required(candidate_row)

    checkpoint_diff = checkpoint_deepdiff(expected_checkpoint, candidate_checkpoint)

    rng_ok, rng_stream_detail = compare_rng_stream(
        rng_stream_channel_required(expected_row),
        rng_stream_channel_required(candidate_row),
    )
    rng_stream = dict(rng_stream_detail or {})
    rng_stream["ok"] = bool(rng_ok)

    entity_presence: BuiltinObject = {}
    entity_diverged = False
    for kind in ENTITY_SAMPLE_KINDS:
        expected_uids = _entity_uid_set(expected_row, kind)
        candidate_uids = _entity_uid_set(candidate_row, kind)
        missing = sorted(expected_uids - candidate_uids)
        extra = sorted(candidate_uids - expected_uids)
        diverged = bool(missing or extra or len(expected_uids) != len(candidate_uids))
        if diverged:
            entity_diverged = True
        entity_presence[kind] = {
            "expected_count": len(expected_uids),
            "candidate_count": len(candidate_uids),
            "missing_uids": missing,
            "extra_uids": extra,
        }
    entity_samples_ok, entity_samples_detail = compare_entity_samples(
        entity_samples_channel_required(expected_row),
        entity_samples_channel_required(candidate_row),
    )
    sim_state_ok, sim_state_detail = compare_sim_state(
        sim_state_channel_required(expected_row),
        sim_state_channel_required(candidate_row),
    )

    diverged = bool(
        checkpoint_diff is not None
        or not bool(rng_stream.get("ok"))
        or entity_diverged
        or not entity_samples_ok
        or not sim_state_ok,
    )

    return to_builtin_object(
        {
            "tick_index": int(tick),
            "diverged": diverged,
            "checkpoint_diff_count": (0 if checkpoint_diff is None else int(checkpoint_diff.diff_count)),
            "checkpoint_diff": (
                None
                if checkpoint_diff is None
                else {
                    "payload": checkpoint_diff.payload,
                    "pretty": checkpoint_diff.pretty,
                }
            ),
            "rng_stream": rng_stream,
            "entity_presence": entity_presence,
            "entity_samples": {
                "ok": bool(entity_samples_ok),
                "detail": entity_samples_detail,
            },
            "sim_state": {
                "ok": bool(sim_state_ok),
                "detail": sim_state_detail,
            },
        },
        field="focus_tick",
    )



from __future__ import annotations

from typing import Annotated, TypeAlias

import msgspec

ROOM_CODE_LENGTH = 4

RoomCode: TypeAlias = Annotated[
    str,
    msgspec.Meta(
        min_length=ROOM_CODE_LENGTH,
        max_length=ROOM_CODE_LENGTH,
        pattern=rf"^[a-z0-9]{{{ROOM_CODE_LENGTH}}}$",
    ),
]


def parse_room_code(value: str) -> RoomCode:
    return msgspec.convert(str(value).strip().lower(), type=RoomCode)


def parse_optional_room_code(value: str | None) -> RoomCode | None:
    text = str(value or "").strip()
    if not text:
        return None
    return parse_room_code(text)


__all__ = ["ROOM_CODE_LENGTH", "RoomCode", "parse_optional_room_code", "parse_room_code"]
